import torch as t
import torch
import torch.nn as nn
import torch.nn.functional as F
import scipy.sparse as sp
import numpy as np
from openhgnn.models.base_model import BaseModel
from openhgnn.models import register_model
import dgl

### HGCL
@register_model('HGCL')
class HGCL(BaseModel):
    def build_model_from_args(args, hg):
        userNum = hg.number_of_nodes('user')
        itemNum = hg.number_of_nodes('item')

        userMat = hg.adj_external(etype=('user', 'distance', 'user'), scipy_fmt='csr')
        itemMat = hg.adj_external(etype=('item', 'distance', 'item'), scipy_fmt='csr')
        uiMat = hg.adj_external(etype=('user+item', 'distance', 'user+item'), scipy_fmt='csr')

        return HGCL(userNum=userNum, itemNum=itemNum, userMat=userMat, itemMat=itemMat, uiMat=uiMat,
                   hide_dim=args.hide_dim, Layers=args.Layers, rank=args.rank, wu1=args.wu1,
                   wu2=args.wu2, wi1=args.wi1, wi2=args.wi2)
    def __init__(self, userNum, itemNum, userMat, itemMat, uiMat, hide_dim, Layers, rank, wu1, wu2, wi1, wi2):
        super(HGCL, self).__init__()
        self.userNum = userNum
        self.itemNum = itemNum
        self.uuMat = userMat
        self.iiMat = itemMat
        self.uiMat = uiMat
        self.hide_dim = hide_dim
        self.LayerNums = Layers
        self.wu1 = wu1
        self.wu2 = wu2
        self.wi1 = wi1
        self.wi2 = wi2

        uimat = self.uiMat[: self.userNum, self.userNum:]
        values = torch.FloatTensor(uimat.tocoo().data)
        indices = np.vstack((uimat.tocoo().row, uimat.tocoo().col))
        i = torch.LongTensor(indices)
        v = torch.FloatTensor(values)
        shape = uimat.tocoo().shape
        uimat1 = torch.sparse.FloatTensor(i, v, torch.Size(shape))
        self.uiadj = uimat1
        self.iuadj = uimat1.transpose(0, 1)

        self.gating_weightub = nn.Parameter(
            torch.FloatTensor(1, hide_dim))
        nn.init.xavier_normal_(self.gating_weightub.data)
        self.gating_weightu = nn.Parameter(
            torch.FloatTensor(hide_dim, hide_dim))
        nn.init.xavier_normal_(self.gating_weightu.data)
        self.gating_weightib = nn.Parameter(
            torch.FloatTensor(1, hide_dim))
        nn.init.xavier_normal_(self.gating_weightib.data)
        self.gating_weighti = nn.Parameter(
            torch.FloatTensor(hide_dim, hide_dim))
        nn.init.xavier_normal_(self.gating_weighti.data)

        self.encoder = nn.ModuleList()
        for i in range(0, self.LayerNums):
            self.encoder.append(GCN_layer())
        self.k = rank
        k = self.k
        self.mlp = MLP(hide_dim, hide_dim * k, hide_dim // 2, hide_dim * k)
        self.mlp1 = MLP(hide_dim, hide_dim * k, hide_dim // 2, hide_dim * k)
        self.mlp2 = MLP(hide_dim, hide_dim * k, hide_dim // 2, hide_dim * k)
        self.mlp3 = MLP(hide_dim, hide_dim * k, hide_dim // 2, hide_dim * k)
        self.meta_netu = nn.Linear(hide_dim * 3, hide_dim, bias=True)
        self.meta_neti = nn.Linear(hide_dim * 3, hide_dim, bias=True)

        self.embedding_dict = nn.ModuleDict({
            'uu_emb': torch.nn.Embedding(userNum, hide_dim).cuda(),
            'ii_emb': torch.nn.Embedding(itemNum, hide_dim).cuda(),
            'user_emb': torch.nn.Embedding(userNum, hide_dim).cuda(),
            'item_emb': torch.nn.Embedding(itemNum, hide_dim).cuda(),
        })

    def init_weight(self, userNum, itemNum, hide_dim):
        initializer = nn.init.xavier_uniform_
        embedding_dict = nn.ParameterDict({
            'user_emb': nn.Parameter(initializer(t.empty(userNum, hide_dim))),
            'item_emb': nn.Parameter(initializer(t.empty(itemNum, hide_dim))),
        })
        return embedding_dict

    def sparse_mx_to_torch_sparse_tensor(self, sparse_mx):
        """Convert a scipy sparse matrix to a torch sparse tensor."""
        if type(sparse_mx) != sp.coo_matrix:
            sparse_mx = sparse_mx.tocoo().astype(np.float32)
        indices = torch.from_numpy(
            np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
        values = torch.from_numpy(sparse_mx.data).float()
        shape = torch.Size(sparse_mx.shape)
        return torch.sparse.FloatTensor(indices, values, shape)

    def metaregular(self, em0, em, adj):
        def row_column_shuffle(embedding):
            corrupted_embedding = embedding[:, torch.randperm(embedding.shape[1])]
            corrupted_embedding = corrupted_embedding[torch.randperm(embedding.shape[0])]
            return corrupted_embedding

        def score(x1, x2):
            x1 = F.normalize(x1, p=2, dim=-1)
            x2 = F.normalize(x2, p=2, dim=-1)
            return torch.sum(torch.multiply(x1, x2), 1)

        user_embeddings = em
        Adj_Norm = t.from_numpy(np.sum(adj, axis=1)).float().cuda()
        adj = self.sparse_mx_to_torch_sparse_tensor(adj)
        edge_embeddings = torch.spmm(adj.cuda(), user_embeddings) / Adj_Norm
        user_embeddings = em0
        graph = torch.mean(edge_embeddings, 0)
        pos = score(user_embeddings, graph)
        neg1 = score(row_column_shuffle(user_embeddings), graph)
        global_loss = torch.mean(-torch.log(torch.sigmoid(pos - neg1)))
        return global_loss

    def self_gatingu(self, em):
        return torch.multiply(em, torch.sigmoid(torch.matmul(em, self.gating_weightu) + self.gating_weightub))

    def self_gatingi(self, em):
        return torch.multiply(em, torch.sigmoid(torch.matmul(em, self.gating_weighti) + self.gating_weightib))

    def metafortansform(self, auxiembedu, targetembedu, auxiembedi, targetembedi):

        # Neighbor information of the target node
        uneighbor = t.matmul(self.uiadj.cuda(), self.ui_itemEmbedding)
        ineighbor = t.matmul(self.iuadj.cuda(), self.ui_userEmbedding)

        # Meta-knowlege extraction
        tembedu = (self.meta_netu(t.cat((auxiembedu, targetembedu, uneighbor), dim=1).detach()))
        tembedi = (self.meta_neti(t.cat((auxiembedi, targetembedi, ineighbor), dim=1).detach()))

        """ Personalized transformation parameter matrix """
        # Low rank matrix decomposition
        metau1 = self.mlp(tembedu).reshape(-1, self.hide_dim, self.k)  # d*k
        metau2 = self.mlp1(tembedu).reshape(-1, self.k, self.hide_dim)  # k*d
        metai1 = self.mlp2(tembedi).reshape(-1, self.hide_dim, self.k)  # d*k
        metai2 = self.mlp3(tembedi).reshape(-1, self.k, self.hide_dim)  # k*d
        meta_biasu = (torch.mean(metau1, dim=0))
        meta_biasu1 = (torch.mean(metau2, dim=0))
        meta_biasi = (torch.mean(metai1, dim=0))
        meta_biasi1 = (torch.mean(metai2, dim=0))
        low_weightu1 = F.softmax(metau1 + meta_biasu, dim=1)
        low_weightu2 = F.softmax(metau2 + meta_biasu1, dim=1)
        low_weighti1 = F.softmax(metai1 + meta_biasi, dim=1)
        low_weighti2 = F.softmax(metai2 + meta_biasi1, dim=1)

        # The learned matrix as the weights of the transformed network
        tembedus = (t.sum(t.multiply((auxiembedu).unsqueeze(-1), low_weightu1),
                          dim=1))  # Equal to a two-layer linear network; Ciao and Yelp data sets are plus gelu activation function
        tembedus = t.sum(t.multiply((tembedus).unsqueeze(-1), low_weightu2), dim=1)
        tembedis = (t.sum(t.multiply((auxiembedi).unsqueeze(-1), low_weighti1), dim=1))
        tembedis = t.sum(t.multiply((tembedis).unsqueeze(-1), low_weighti2), dim=1)
        transfuEmbed = tembedus
        transfiEmbed = tembedis
        return transfuEmbed, transfiEmbed

    def forward(self, iftraining, uid, iid, norm=1):

        item_index = np.arange(0, self.itemNum)
        user_index = np.arange(0, self.userNum)
        ui_index = np.array(user_index.tolist() + [i + self.userNum for i in item_index])

        # Initialize Embeddings
        userembed0 = self.embedding_dict['user_emb'].weight
        itemembed0 = self.embedding_dict['item_emb'].weight
        uu_embed0 = self.self_gatingu(userembed0) # e0uu
        ii_embed0 = self.self_gatingi(itemembed0) # e0ii
        self.ui_embeddings = t.cat([userembed0, itemembed0], 0) # e0ui
        self.all_user_embeddings = [uu_embed0]
        self.all_item_embeddings = [ii_embed0]
        self.all_ui_embeddings = [self.ui_embeddings]
        # Encoder
        for i in range(len(self.encoder)):
            layer = self.encoder[i]
            if i == 0:
                userEmbeddings0 = layer(uu_embed0, self.uuMat, user_index)
                itemEmbeddings0 = layer(ii_embed0, self.iiMat, item_index)
                uiEmbeddings0 = layer(self.ui_embeddings, self.uiMat, ui_index)
            else:
                userEmbeddings0 = layer(userEmbeddings, self.uuMat, user_index)
                itemEmbeddings0 = layer(itemEmbeddings, self.iiMat, item_index)
                uiEmbeddings0 = layer(uiEmbeddings, self.uiMat, ui_index)

            # Aggregation of message features across the two related views in the middle layer then fed into the next layer
            self.ui_userEmbedding0, self.ui_itemEmbedding0 = t.split(uiEmbeddings0, [self.userNum, self.itemNum])
            userEd = (userEmbeddings0 + self.ui_userEmbedding0) / 2.0
            itemEd = (itemEmbeddings0 + self.ui_itemEmbedding0) / 2.0
            userEmbeddings = userEd
            itemEmbeddings = itemEd
            uiEmbeddings = torch.cat([userEd, itemEd], 0)
            if norm == 1:
                norm_embeddings = F.normalize(userEmbeddings0, p=2, dim=1)
                self.all_user_embeddings += [norm_embeddings]
                norm_embeddings = F.normalize(itemEmbeddings0, p=2, dim=1)
                self.all_item_embeddings += [norm_embeddings]
                norm_embeddings = F.normalize(uiEmbeddings0, p=2, dim=1)
                self.all_ui_embeddings += [norm_embeddings]
            else:
                self.all_user_embeddings += [userEmbeddings]
                self.all_item_embeddings += [norm_embeddings]
                self.all_ui_embeddings += [norm_embeddings]
        self.userEmbedding = t.stack(self.all_user_embeddings, dim=1)
        self.userEmbedding = t.mean(self.userEmbedding, dim=1)
        self.itemEmbedding = t.stack(self.all_item_embeddings, dim=1)
        self.itemEmbedding = t.mean(self.itemEmbedding, dim=1)
        self.uiEmbedding = t.stack(self.all_ui_embeddings, dim=1)
        self.uiEmbedding = t.mean(self.uiEmbedding, dim=1)
        self.ui_userEmbedding, self.ui_itemEmbedding = t.split(self.uiEmbedding, [self.userNum, self.itemNum])

        # Personalized Transformation of Auxiliary Domain Features
        metatsuembed, metatsiembed = self.metafortansform(self.userEmbedding, self.ui_userEmbedding, self.itemEmbedding,
                                                          self.ui_itemEmbedding)
        self.userEmbedding = self.userEmbedding + metatsuembed
        self.itemEmbedding = self.itemEmbedding + metatsiembed

        # Regularization: the constraint of transformed reasonableness
        metaregloss = 0
        if iftraining == True:
            self.reg_lossu = self.metaregular((self.ui_userEmbedding[uid.cpu().numpy()]), (self.userEmbedding),
                                              self.uuMat[uid.cpu().numpy()])
            self.reg_lossi = self.metaregular((self.ui_itemEmbedding[iid.cpu().numpy()]), (self.itemEmbedding),
                                              self.iiMat[iid.cpu().numpy()])
            metaregloss = (self.reg_lossu + self.reg_lossi) / 2.0
        return self.userEmbedding, self.itemEmbedding, (
                    self.wu1 * self.ui_userEmbedding + self.wu2 * self.userEmbedding), (
                    self.wi1 * self.ui_itemEmbedding + self.wi2 * self.itemEmbedding), self.ui_userEmbedding, self.ui_itemEmbedding, metaregloss


class GCN_layer(nn.Module):
    def __init__(self):
        super(GCN_layer, self).__init__()

    def sparse_mx_to_torch_sparse_tensor(self, sparse_mx):
        """Convert a scipy sparse matrix to a torch sparse tensor."""
        if type(sparse_mx) != sp.coo_matrix:
            sparse_mx = sparse_mx.tocoo().astype(np.float32)
        indices = torch.from_numpy(
            np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
        values = torch.from_numpy(sparse_mx.data).float()
        shape = torch.Size(sparse_mx.shape)
        return torch.sparse.FloatTensor(indices, values, shape)

    def normalize_adj(self, adj):
        """Symmetrically normalize adjacency matrix."""
        adj = sp.coo_matrix(adj)
        rowsum = np.array(adj.sum(1))
        d_inv_sqrt = np.power(rowsum, -0.5).flatten()
        d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.
        d_mat_inv_sqrt = sp.diags(d_inv_sqrt)
        return (d_mat_inv_sqrt).dot(adj).dot(d_mat_inv_sqrt).tocoo()

    def forward(self, features, Mat, index):
        subset_Mat = Mat
        subset_features = features
        subset_Mat = self.normalize_adj(subset_Mat)
        subset_sparse_tensor = self.sparse_mx_to_torch_sparse_tensor(subset_Mat).cuda()
        out_features = torch.spmm(subset_sparse_tensor, subset_features)
        new_features = torch.empty(features.shape).cuda()
        new_features[index] = out_features
        dif_index = np.setdiff1d(torch.arange(features.shape[0]), index)
        new_features[dif_index] = features[dif_index]
        return new_features


class MLP(torch.nn.Module):
    def __init__(self, input_dim, feature_dim, hidden_dim, output_dim,
                 feature_pre=True, layer_num=2, dropout=True, **kwargs):
        super(MLP, self).__init__()
        self.feature_pre = feature_pre
        self.layer_num = layer_num
        self.dropout = dropout
        if feature_pre:
            self.linear_pre = nn.Linear(input_dim, feature_dim, bias=True)
        else:
            self.linear_first = nn.Linear(input_dim, hidden_dim)
        self.linear_hidden = nn.ModuleList([nn.Linear(hidden_dim, hidden_dim) for i in range(layer_num - 2)])
        self.linear_out = nn.Linear(feature_dim, output_dim, bias=True)

    def forward(self, data):
        x = data
        if self.feature_pre:
            x = self.linear_pre(x)
        prelu = nn.PReLU().cuda()
        x = prelu(x)
        for i in range(self.layer_num - 2):
            x = self.linear_hidden[i](x)
            x = F.tanh(x)
            if self.dropout:
                x = F.dropout(x, training=self.training)
        x = self.linear_out(x)
        x = F.normalize(x, p=2, dim=-1)
        return x























import torch
import torch.nn as nn
from torch.autograd import Variable
from . import BaseModel, register_model
import torch.nn.functional as F


@register_model('MeiREC')
class MeiREC(BaseModel):
    r"""
        MeiREC from paper `Metapath-guided Heterogeneous Graph Neural Network for
        Intent Recommendation <https://dl.acm.org/doi/abs/10.1145/3292500.3330673>`__
        in KDD_2019.

        `Code from author <https://github.com/googlebaba/KDD2019-MEIRec>`__.

        We leverage metapaths to obtain different-step neighbors of an object, and the embeddings of us
        ers and queries are the aggregation of their neighbors under different metapaths.And we propose
        to represent the queries and items with a small number of term embeddings.we need to learn the
        term embeddings, rather than all object embeddings. This method is able to significantly reduc
        e the number of parameters.

        Parameters
        ----------
        user_seq_length : int
            Number for process dataset.
        ...
        batch_num : int
            Number of batch.
        weight_decay : float
            Number of weight_decay.
        lr : float
            learning rate.
        train_epochs : int
            Number of train epoch.
        -----------
    """

    @classmethod
    def build_model_from_args(cls, config):
        return cls(config)

    def __init__(self, config):
        super().__init__()

        self.model = Model(config)

    def forward(self, *args):
        return self.model(*args)

    def extra_loss(self):
        pass


class Model(nn.Module):

    def __init__(self, args):
        super().__init__()
        self.args = args

        # model params
        self.user_seq_length = 15
        self.user_item_term_length = 10
        self.user_query_term_length = 10
        self.query_length = 10
        self.query_topcate_length = 3
        self.query_leafcate_length = 3
        self.embed_size_word = 64

        self._generate_model_layer()

        # drop out prob
        self.keep_prob = 0.8

    def set_mode(self, mode="train"):
        self.keep_prob = 0.8 if mode == "train" else 1.0

    def _generate_model_layer(self):

        self._word_embed = nn.Parameter(
            Variable(
                torch.Tensor(self.args.vocab,
                             self.embed_size_word),
                requires_grad=True,
            ))
        # self.register_parameter('word_embed', nn.Parameter(self._word_embed))
        # lstms
        # h,c for lstm weights, w b for mlp weights, bias
        (
            self.user_word_lstm,
            self.h_1,
            self.c_1,
            self.w_l1,
            self.b_l1,
            self.loss1,
        ) = self._rnn_lstm(1, self.args.batch_num, 64, 64)

        (
            self.user_item_query_lstm,
            self.h_2,
            self.c_2,
            self.w_l2,
            self.b_l2,
            self.loss2,
        ) = self._rnn_lstm(1, self.args.batch_num, 64, 64)

        (
            self.user_query_item_lstm,
            self.h_3,
            self.c_3,
            self.w_l3,
            self.b_l3,
            self.loss3,
        ) = self._rnn_lstm(1, self.args.batch_num, 64, 64)

        (
            self.user_query_seq_lstm,
            self.h_4,
            self.c_4,
            self.w_l4,
            self.b_l4,
            self.loss4,
        ) = self._rnn_lstm(1, self.args.batch_num, 64, 64)

        # cnn
        self.query_item_query_cnn, self.conv_w1, self.loss5 = self._cnn(64)
        self.query_item_query_linear = nn.Linear(12 * 2 * 1, 64)
        self.query_item_query_relu = nn.ReLU()

        self.query_user_item_cnn, self.conv_w2, self.loss6 = self._cnn(64)
        self.query_user_item_linear = nn.Linear(12 * 2 * 1, 64)
        self.query_user_item_relu = nn.ReLU()

        # weights & bias
        # wide feats, mlp layers for wide feats, 81 for static feats length
        self.wide_feat_w, self.loss7 = self.get_weights_variables(
            [64, 81], self.args.weight_decay)
        self.wide_feat_b = self.get_bias_variables(64)
        # concat query weights, 64 * 7 for concat feats len
        self.concat_query_w, self.loss8 = self.get_weights_variables(
            [64, 64 * 7], self.args.weight_decay)
        self.concat_query_b = self.get_bias_variables(64)
        # concat_query_user_wide,  concat query and
        # user wide infos, then len: 128
        self.concat_query_user_wide_w, self.loss9 = self.get_weights_variables(
            [64, 128], self.args.weight_decay)
        self.concat_query_user_wide_b = self.get_bias_variables(64)
        # deep_wide_feat, the last layer mlp => predict val
        self.deep_wide_feat_w, self.loss10 = self.get_weights_variables(
            [1, 64], self.args.weight_decay)
        self.deep_wide_feat_b = self.get_bias_variables(1)

    @property
    def regular_loss(self):
        return (
            self.get_regular_loss(self.w_l1) +
            self.get_regular_loss(self.w_l2) +
            self.get_regular_loss(self.w_l3) +
            self.get_regular_loss(self.w_l4) +
            self.get_regular_loss(self.wide_feat_w) +
            self.get_regular_loss(self.concat_query_w) +
            self.get_regular_loss(self.concat_query_user_wide_w) +
            self.get_regular_loss(self.deep_wide_feat_w))

    def get_regular_loss(self, params):

        return torch.sum(torch.pow(params, 2)) / 2 * self.args.weight_decay

    def get_weights_variables(self, shape, weight_decay, trainable=True):

        params = nn.Parameter(
            Variable(torch.Tensor(*shape), requires_grad=trainable))
        if weight_decay == 0:
            regular_loss = 0.0
        else:
            # xavier initializer, need params demension >= 2
            nn.init.xavier_uniform_(params, gain=1.0)
            # l2 regular loss for weights
            regular_loss = torch.sum(torch.pow(params, 2)) / 2 * weight_decay
        return params, regular_loss

    def get_bias_variables(self, size, trainable=True):
        params = nn.Parameter(
            Variable(torch.Tensor(size, 1), requires_grad=trainable))
        # constant initialize of
        nn.init.constant_(params, 0.0)
        return params

    def _rnn_lstm(self, layers_num, batches_num, features_num, hidden_size):
        # LstmCell: features * hidden_size * layers_num
        lstm = nn.LSTM(features_num, hidden_size, layers_num, bias=True)
        # weights: num_layers * batch * hidden size
        h_0, regular_loss_h0 = self.get_weights_variables(
            [layers_num, batches_num, hidden_size], self.args.weight_decay)
        c_0, regular_loss_c0 = self.get_weights_variables(
            [layers_num, batches_num, hidden_size], self.args.weight_decay)

        w_l, regular_loss_wl = self.get_weights_variables(
            [hidden_size, hidden_size], self.args.weight_decay)
        b_l = self.get_bias_variables(hidden_size)
        # hidden_layer = torch.tanh(torch.matmul(w_l, lstm_cell.T) + b_l)
        loss_total = regular_loss_h0 + regular_loss_c0 + regular_loss_wl
        # output: 64 * batch_num
        return lstm, h_0, c_0, w_l, b_l, loss_total

    def _cnn(self, features_num):
        # x: batch * channels * h * w
        # x_input = input.view(-1, 1, nums, features_num)
        conv_w, regular_loss_w = self.get_weights_variables(
            [12, 1, 2, features_num], self.args.weight_decay)
        conv_b, _ = self.get_weights_variables([12], 0)
        # default stride 1, padding valid
        conv = nn.Conv2d(1, 12, kernel_size=(2, features_num), stride=1, padding='valid')
        return conv, conv_w, regular_loss_w

    def forward(self, x):
        # x:[m, batch_size];  912*512
        wide_feat = x[:81, :]
        user_item_seq = x[81:276, :]
        query_feat = x[276:292, :]
        user_query_seq = x[292:462, :]
        query_item_query = x[462:562, :]
        user_query_item = x[562:662, :]
        user_item_query = x[662:812, :]
        query_user_item = x[812:, :]

        # query embedding, query terms 10 * batch
        query_terms, query_topcate, query_leafcate = torch.split(
            query_feat,
            [
                self.query_length,
                self.query_topcate_length,
                self.query_leafcate_length,
            ],
            0,
        )
        # look up in word embedding's dict [280000, 64] => 10 * batch * 64

        inputs_query_raw = torch.nn.functional.embedding(query_terms.to(torch.int64), self._word_embed)
        input_num = (torch.sum(torch.sign(query_terms))
                     if torch.sum(torch.sign(query_terms)) > 1 else
                     torch.tensor(1))
        # get query word to vec sum 64 * batchsize
        self.query_w2v_sum = torch.mean(
            inputs_query_raw[:int(input_num.item())], 0).T        #64*512

        # user word embedding
        raw_word_embedding_list = torch.split(user_item_seq[-13 * 5:, :],
                                              [13] * 5, 0)
        step_embedding_list = []
        for raw_word_embed in raw_word_embedding_list:
            item_terms, item_topcate, item_leafcate, time_delta = torch.split(
                raw_word_embed, [self.user_item_term_length, 1, 1, 1],
                0)

            step_embedding = torch.nn.functional.embedding(item_terms.to(torch.int64), self._word_embed)

            input_num = (torch.sum(torch.sign(item_terms))
                         if torch.sum(torch.sign(item_terms)) > 1 else
                         torch.tensor(1))
            # step_avg_embedding: batchsize * 64
            step_avg_embedding = torch.mean(
                step_embedding[:int(input_num.item())], 0)
            # append a new axis in the first
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        # step_embedding vec: 5 * batchsize * 64
        step_embedding_vec = torch.cat(step_embedding_list, 0)

        lstm_cells, lstm_hiddens = self.user_word_lstm(step_embedding_vec,
                                                       (self.h_1, self.c_1))
        self.user_item_term_lstm_output = torch.tanh(
            torch.matmul(self.w_l1, lstm_cells[-1].T) + self.b_l1)

        # user_item_query_embedding
        raw_user_item_query_embedding_list = torch.split(
            user_item_query[:10 * 5, :], [10] * 5, 0)
        step_embedding_list = []
        for raw_user_item_embed in raw_user_item_query_embedding_list:
            item_terms = raw_user_item_embed[:5]

            step_embedding = torch.nn.functional.embedding(item_terms.to(torch.int64), self._word_embed)

            input_num = (torch.sum(torch.sign(item_terms))
                         if torch.sum(torch.sign(item_terms)) > 1 else
                         torch.tensor(1))
            step_avg_embedding = torch.mean(
                step_embedding[:int(input_num.item())], 0)
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        step_embedding_vec = torch.cat(step_embedding_list, 0)
        lstm_cells, lstm_hiddens = self.user_item_query_lstm(
            step_embedding_vec, (self.h_2, self.c_2))
        self.user_item_query_term_lstm_output = torch.tanh(
            torch.matmul(self.w_l2, lstm_cells[-1].T) + self.b_l2)

        # user_query_item embedding
        raw_user_query_item_embedding_list = torch.split(
            user_query_item[-10 * 5:, :], [10] * 5, 0)
        step_embedding_list = []
        for raw_user_item_embed in raw_user_query_item_embedding_list:
            item_terms = raw_user_item_embed[:5]

            step_embedding = torch.nn.functional.embedding(item_terms.to(torch.int64), self._word_embed)

            input_num = (torch.sum(torch.sign(item_terms))
                         if torch.sum(torch.sign(item_terms)) > 1 else
                         torch.tensor(1))
            step_avg_embedding = torch.mean(
                step_embedding[:int(input_num.item())], 0)
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        step_embedding_vec = torch.cat(step_embedding_list, 0)
        lstm_cells, lstm_hiddens = self.user_query_item_lstm(
            step_embedding_vec, (self.h_3, self.c_3))
        self.user_query_item_term_lstm_output = torch.tanh(
            torch.matmul(self.w_l3, lstm_cells[-1].T) + self.b_l3)

        # query_item_query embed
        raw_query_item_query_embedding_list = torch.split(
            query_item_query[-10 * 5:, :], [10] * 5, 0)
        step_embedding_list = []
        for raw_user_item_embed in raw_query_item_query_embedding_list:
            query_terms = raw_user_item_embed[:5]

            step_embedding = torch.nn.functional.embedding(query_terms.to(torch.int64), self._word_embed)

            input_num = (torch.sum(torch.sign(query_terms))
                         if torch.sum(torch.sign(query_terms)) > 1 else
                         torch.tensor(1))
            step_avg_embedding = torch.mean(                       # 512 * 64
                step_embedding[:int(input_num.item())], 0)
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        step_embedding_vec = torch.cat(step_embedding_list, 0)     # 5 * 512 * 64

        convd = self.query_item_query_cnn(
            torch.transpose(step_embedding_vec, 0, 1).unsqueeze(1))
        convd_active = F.relu(convd)
        pooled = F.max_pool2d(convd_active, (2, 1), stride=2)
        pooled = torch.transpose(pooled, 1, 2)
        pooled = torch.transpose(pooled, 2, 3)
        pool_flat = pooled.reshape(-1, 2 * 1 * 12)
        self.query_item_query_cnn_output = self.query_item_query_relu(
            self.query_item_query_linear(pool_flat)).T

        # query_user_item embedd
        raw_query_user_item_embedding_list = torch.split(
            query_user_item[-10 * 5:, :], [10] * 5, 0)
        step_embedding_list = []
        for raw_user_item_embed in raw_query_user_item_embedding_list:
            item_terms = raw_user_item_embed[:5]

            step_embedding = torch.nn.functional.embedding(item_terms.to(torch.int64), self._word_embed)

            input_num = (torch.sum(torch.sign(item_terms))
                         if torch.sum(torch.sign(item_terms)) > 1 else
                         torch.tensor(1))
            step_avg_embedding = torch.mean(
                step_embedding[:int(input_num.item())], 0)
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        step_embedding_vec = torch.cat(step_embedding_list, 0)

        convd = self.query_user_item_cnn(
            torch.transpose(step_embedding_vec, 0, 1).unsqueeze(1))
        convd_active = F.relu(convd)
        pooled = F.max_pool2d(convd_active, (2, 1), stride=2)
        pooled = torch.transpose(pooled, 1, 2)
        pooled = torch.transpose(pooled, 2, 3)
        pool_flat = pooled.reshape(-1, 2 * 1 * 12)
        self.query_user_item_cnn_output = self.query_user_item_relu(
            self.query_user_item_linear(pool_flat)).T

        #  user_query_seq embed
        raw_user_query_seq_embedding_list = torch.split(
            user_query_seq[-17 * 5:, :], [17] * 5, 0)
        step_embedding_list = []
        for raw_user_item_embed in raw_user_query_seq_embedding_list[::-1]:
            (
                query_terms,
                query_topcate,
                query_leafcate,
                time_delta,
            ) = torch.split(raw_user_item_embed,
                            [self.user_query_term_length, 3, 3, 1], 0)

            step_embedding = torch.nn.functional.embedding(query_terms.to(torch.int64), self._word_embed)
            input_num = (torch.sum(torch.sign(query_terms))
                         if torch.sum(torch.sign(query_terms)) > 1 else
                         torch.tensor(1))
            step_avg_embedding = torch.mean(
                step_embedding[:int(input_num.item())], 0)
            step_embedding_list.append(step_avg_embedding.unsqueeze(0))
        step_embedding_vec = torch.cat(step_embedding_list, 0)
        lstm_cells, lstm_hiddens = self.user_query_seq_lstm(
            step_embedding_vec, (self.h_4, self.c_4))
        self.user_query_term_lstm_output = torch.tanh(
            torch.matmul(self.w_l4, lstm_cells[-1].T) + self.b_l4)

        # wide connected, connect static params
        # wide feat hidden output: 64 * batch_size
        self.wide_hidden_layer1 = torch.tanh(                        # [64, 81] * [81, 1]
            F.dropout(
                torch.matmul(self.wide_feat_w, wide_feat.float()) +
                self.wide_feat_b,
                self.keep_prob
            ))
        concat_seq = [
            self.user_item_term_lstm_output,
            self.user_query_term_lstm_output,
            self.query_w2v_sum,
            self.user_item_query_term_lstm_output,
            self.user_query_item_term_lstm_output,
            self.query_item_query_cnn_output,
            self.query_user_item_cnn_output,
        ]
        qu_term_concat = F.dropout(torch.cat(concat_seq, 0), self.keep_prob)    # [64*7, 1]
        # concat user query feats: 64 * batch_size
        self.qu_term_hidden_layer1 = torch.tanh(
            F.dropout(
                torch.matmul(self.concat_query_w, qu_term_concat) +    #聚合特征信息 [64, 64*7] * [64*7, 1]
                self.concat_query_b,
                self.keep_prob
            ))
        # concat feats and static datas: 128 * batch_size
        deep_wide_concat = torch.cat(                  # [64, 1] || [64, 1] -> [128, 1]
            [self.qu_term_hidden_layer1, self.wide_hidden_layer1], 0)
        # mlp for wide concat data : 64 * batch_size
        self.dw_hidden_layer0 = torch.tanh(            # [64, 128] * [128 * 1]  mlp
            F.dropout(
                torch.matmul(self.concat_query_user_wide_w, deep_wide_concat) +
                self.concat_query_user_wide_b,
                self.keep_prob
            ))
        self.dw_hidden_layer1 = torch.sigmoid(         # [1, 64] * [64, 1] -> 1
            torch.matmul(self.deep_wide_feat_w, self.dw_hidden_layer0) +
            self.deep_wide_feat_b)
        self.predict_labels = self.dw_hidden_layer1.squeeze()
        return self.predict_labels




import dgl
import torch
import torch.nn as nn
import dgl.function as Fn
import torch.nn.functional as F

from dgl.ops import edge_softmax
from dgl.nn.pytorch import TypedLinear
from ..utils import to_hetero_feat
from . import BaseModel, register_model

@register_model('HetSANN')
class HetSANN(BaseModel):
    r"""
    This is a model HetSANN from `An Attention-Based Graph Neural Network for Heterogeneous Structural Learning
    <https://arxiv.org/abs/1912.10832>`__

    It contains the following part:

    Apply a linear transformation:
    
    .. math::
       h^{(l+1, m)}_{\phi(j),i} = W^{(l+1, m)}_{\phi(j),\phi(i)} h^{(l)}_i \quad (1)
    
    And return the new embeddings.
    
    You may refer to the paper HetSANN-Section 2.1-Type-aware Attention Layer-(1)

    Aggregation of Neighborhood:
    
    Computing the attention coefficient:
    
    .. math::
       o^{(l+1,m)}_e = \sigma(f^{(l+1,m)}_r(h^{(l+1, m)}_{\phi(j),j}, h^{(l+1, m)}_{\phi(j),i})) \quad (2)
        
    .. math::
       f^{(l+1,m)}_r(e) = [h^{(l+1, m)^T}_{\phi(j),j}||h^{(l+1, m)^T}_{\phi(j),i}]a^{(l+1, m)}_r ] \quad (3)
    
    .. math::
       \alpha^{(l+1,m)}_e = exp(o^{(l+1,m)}_e) / \sum_{k\in \varepsilon_j} exp(o^{(l+1,m)}_k) \quad (4)
    
    Getting new embeddings with multi-head and residual
    
    .. math::
       h^{(l + 1, m)}_j = \sigma(\sum_{e = (i,j,r)\in \varepsilon_j} \alpha^{(l+1,m)}_e h^{(l+1, m)}_{\phi(j),i}) \quad (5)
    
    Multi-heads:
    
    .. math::
       h^{(l+1)}_j = \parallel^M_{m = 1}h^{(l + 1, m)}_j \quad (6)
    
    Residual:
    
    .. math::
       h^{(l+1)}_j = h^{(l)}_j + \parallel^M_{m = 1}h^{(l + 1, m)}_j \quad (7)
    
    Parameters
    ----------
    num_heads: int
        the number of heads in the attention computing
    num_layers: int
        the number of layers we used in the computing
    in_dim: int
        the input dimension
    num_classes: int
        the number of the output classes
    num_etypes: int
        the number of the edge types
    dropout: float
        the dropout rate
    negative_slope: float
        the negative slope used in the LeakyReLU
    residual: boolean
        if we need the residual operation
    ntype: list
        the list of node type
    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(
            args.num_heads,
            args.num_layers,
            args.hidden_dim,
            args.out_dim,
            hg.ntypes,
            len(hg.etypes),
            args.dropout,
            args.slope,
            args.residual,
            )
    
    def __init__(self, num_heads, num_layers, in_dim, num_classes,
                ntypes, num_etypes, dropout, negative_slope, residual):
        super(HetSANN, self).__init__()
        self.num_layers = num_layers
        self.ntypes = ntypes
        # self.dropout = nn.Dropout(dropout)
        self.residual = residual
        self.activation = F.elu
        
        self.het_layers = nn.ModuleList()
        
        # input projection
        self.het_layers.append(
            HetSANNConv(
                num_heads,
                in_dim,
                in_dim // num_heads,
                num_etypes,
                dropout,
                negative_slope,
                False,
                self.activation,
            )
        )

        # hidden layer
        for i in range(1, num_layers - 1):
            self.het_layers.append(
                HetSANNConv(
                    num_heads,
                    in_dim,
                    in_dim // num_heads,
                    num_etypes,
                    dropout,
                    negative_slope,
                    residual,
                    self.activation
                )
            )

        # output projection
        self.het_layers.append(
            HetSANNConv(
                1,
                in_dim,
                num_classes,
                num_etypes,
                dropout,
                negative_slope,
                residual,
                None,
            )
        )

    def forward(self, hg, h_dict):
        """
        The forward part of the HetSANN.
        
        Parameters
        ----------
        hg : object
            the dgl heterogeneous graph
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after the output projection.
        """
        if hasattr(hg, 'ntypes'):
            with hg.local_scope():
                # input layer and hidden layers
                hg.ndata['h'] = h_dict
                g = dgl.to_homogeneous(hg, ndata = 'h')
                h = g.ndata['h']
                for i in range(self.num_layers - 1):
                    h = self.het_layers[i](g, h, g.ndata['_TYPE'], g.edata['_TYPE'], True)

                # output layer
                h = self.het_layers[-1](g, h, g.ndata['_TYPE'], g.edata['_TYPE'], True)

                h_dict = to_hetero_feat(h, g.ndata['_TYPE'], self.ntypes)
            
        else:
            # for minibatch training, input h_dict is a tensor
            h = h_dict
            for layer, block in zip(self.het_layers, hg):
                h = layer(block, h, block.ndata['_TYPE']['_N'], block.edata['_TYPE'], presorted=False)
            h_dict = to_hetero_feat(h, block.ndata['_TYPE']['_N'][:block.num_dst_nodes()], self.ntypes)

        return h_dict
    
    @property
    def to_homo_flag(self):
        return True

class HetSANNConv(nn.Module):
    """
    The HetSANN convolution layer.

    Parameters
    ----------
    num_heads: int
        the number of heads in the attention computing
    in_dim: int
        the input dimension of the features
    hidden_dim: int
        the hidden dimension of the features
    num_etypes: int
        the number of the edge types
    dropout: float
        the dropout rate
    negative_slope: float
        the negative slope used in the LeakyReLU
    residual: boolean
        if we need the residual operation
    activation: str
        the activation function
    """
    def __init__(self, num_heads, in_dim, hidden_dim, num_etypes,
                 dropout, negative_slope, residual, activation):
        super(HetSANNConv, self).__init__()
        self.num_heads = num_heads
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim

        self.W = TypedLinear(self.in_dim, self.hidden_dim * self.num_heads, num_etypes)
        # self.W_out = TypedLinear(hidden_dim * num_heads, num_classes, num_etypes)

        # self.W_hidden = nn.ModuleDict()
        # self.W_out = nn.ModuleDict()
        
        # for etype in etypes:
        #     self.W_hidden[etype] = nn.Linear(in_dim, hidden_dim * num_heads)
        
        # for etype in etypes:
        #     self.W_out[etype] = nn.Linear(hidden_dim * num_heads, num_classes)
        
        self.a_l = TypedLinear(self.hidden_dim * self.num_heads, self.hidden_dim * self.num_heads, num_etypes)
        self.a_r = TypedLinear(self.hidden_dim * self.num_heads, self.hidden_dim * self.num_heads, num_etypes)
        
        self.dropout = nn.Dropout(dropout)
        self.leakyrelu = nn.LeakyReLU(negative_slope)
        if residual:
            self.residual = nn.Linear(in_dim, self.hidden_dim * num_heads)
        else:
            self.register_buffer("residual", None)
            
        self.activation = activation
        
        
    def forward(self, g, x, ntype, etype, presorted = False):
        """
        The forward part of the HetSANNConv.

        Parameters
        ----------
        g : object
            the dgl homogeneous graph
        x: tensor
            the original features of the graph
        ntype: tensor
            the node type of the graph
        etype: tensor
            the edge type of the graph
        presorted: boolean
            if the ntype and etype are preordered, default: ``False``
            
        Returns
        -------
        tensor
            The embeddings after aggregation.
        """
        # formula (1)
        g.srcdata['h'] = x
        g.apply_edges(Fn.copy_u('h', 'm'))
        h = g.edata['m']
        feat = self.W(h, etype, presorted)
        h = self.dropout(feat)
        g.edata['m'] = h
        h = h.view(-1, self.num_heads, self.hidden_dim)

        # formula (2) (3) (4)
        h_l = self.a_l(h.view(-1, self.num_heads * self.hidden_dim), etype, presorted) \
        .view(-1, self.num_heads, self.hidden_dim).sum(dim = -1)

        h_r = self.a_r(h.view(-1, self.num_heads * self.hidden_dim), etype, presorted) \
        .view(-1, self.num_heads, self.hidden_dim).sum(dim = -1)
        
        attention = self.leakyrelu(h_l + h_r)
        attention = edge_softmax(g, attention)
        
        # formula (5) (6)
        with g.local_scope():
            h = h.permute(0, 2, 1).contiguous()
            g.edata['alpha'] = h @ attention.reshape(-1, self.num_heads, 1)
            
            g.update_all(Fn.copy_e('m', 'w'),Fn.sum('w', 'emb'))
            h_output = g.dstdata['emb']
            # h_prime = []
            # h = h.permute(1, 0, 2).contiguous()
            # for i in range(self.num_heads):
            #     g.edata['alpha'] = attention[:, i]
            #     g.srcdata.update({'emb': h[i]})
            #     g.update_all(Fn.u_mul_e('emb', 'alpha', 'm'),
            #                  Fn.sum('m', 'emb'))
            #     h_prime.append(g.ndata['emb'])
            # h_output = torch.cat(h_prime, dim=1)

        # formula (7)
        if g.is_block:
            x = x[:g.num_dst_nodes()] 
        if self.residual:
            res = self.residual(x)
            h_output += res
        
        if self.activation is not None:
            h_output = self.activation(h_output)       
        
        return h_output



import multiprocessing
from functools import partial, reduce
import time
import numpy as np
import torch
import dgl


class NeighborSampler(object):
    def __init__(self, g, num_fanouts):
        self.g = g
        self.num_fanouts = num_fanouts

    def sample(self, pairs):
        heads, tails, types = zip(*pairs)
        seeds, head_invmap = torch.unique(torch.LongTensor(heads), return_inverse=True)
        blocks = []
        for fanout in reversed(self.num_fanouts):
            sampled_graph = dgl.sampling.sample_neighbors(self.g, seeds, fanout)
            sampled_block = dgl.to_block(sampled_graph, seeds)
            seeds = sampled_block.srcdata[dgl.NID]
            blocks.insert(0, sampled_block)
        return (
            blocks,
            torch.LongTensor(head_invmap),
            torch.LongTensor(tails),
            torch.LongTensor(types),
        )


def generate_pairs(all_walks, window_size, num_workers):
    # for each node, choose the first neighbor and second neighbor of it to form pairs
    # Get all worker processes
    start_time = time.time()
    print("We are generating pairs with {} cores.".format(num_workers))

    # Start all worker processes
    pool = multiprocessing.Pool(processes=num_workers)
    pairs = []
    skip_window = window_size // 2
    for layer_id, walks in enumerate(all_walks):
        block_num = len(walks) // num_workers
        if block_num > 0:
            walks_list = [
                walks[i * block_num: min((i + 1) * block_num, len(walks))]
                for i in range(num_workers)
            ]
        else:
            walks_list = [walks]
        tmp_result = pool.map(
            partial(
                generate_pairs_parallel, skip_window=skip_window, layer_id=layer_id
            ),
            walks_list,
        )
        pairs += reduce(lambda x, y: x + y, tmp_result)

    pool.close()
    end_time = time.time()
    print("Generate pairs end, use {}s.".format(end_time - start_time))
    return np.array([list(pair) for pair in set(pairs)])


def generate_pairs_parallel(walks, skip_window=None, layer_id=None):
    pairs = []
    for walk in walks:
        walk = walk.tolist()
        for i in range(len(walk)):
            for j in range(1, skip_window + 1):
                if i - j >= 0:
                    pairs.append((walk[i], walk[i - j], layer_id))
                if i + j < len(walk):
                    pairs.append((walk[i], walk[i + j], layer_id))
    return pairs



import os.path
from .config import Config
from .utils import set_random_seed, set_best_config, Logger
from .trainerflow import build_flow
from .auto import hpo_experiment

__all__ = ['Experiment']


class Experiment(object):
    r"""Experiment.

    Parameters
    ----------
    model : str or nn.Module
        Name of the model or a hetergenous gnn model provided by the user.
    dataset : str or DGLDataset
        Name of the model or a DGLDataset provided by the user.
    use_best_config: bool
        Whether to load the best config of specific models and datasets. Default: False
    load_from_pretrained : bool
        Whether to load the model from the checkpoint. Default: False
    hpo_search_space :
        Search space for hyperparameters.
    hpo_trials : int
        Number of trials for hyperparameter search.
    Examples
    --------
    >>> experiment = Experiment(model='RGCN', dataset='imdb4GTN', task='node_classification', gpu=-1)
    >>> experiment.run()
    """

    default_conf_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'config.ini')
    specific_trainerflow = {
        'HetGNN': 'hetgnntrainer',
        'HGNN_AC': 'node_classification_ac',
        'NSHE': 'nshetrainer',
        'HeCo': 'HeCo_trainer',
        'DMGI': 'DMGI_trainer',
        'KGCN': 'kgcntrainer',
        'Metapath2vec': 'mp2vec_trainer',
        'HERec': 'herec_trainer',
        'SLiCE': 'slicetrainer',
        'HeGAN': 'HeGAN_trainer',
        'HDE': 'hde_trainer',
        'GATNE-T': 'GATNE_trainer',
        'TransE': 'TransX_trainer',
        'TransH': 'TransX_trainer',
        'TransR': 'TransX_trainer',
        'TransD': 'TransX_trainer',
        'GIE': 'TransX_trainer',
        'HAN': {
            'node_classification': 'han_nc_trainer',
            'link_prediction': 'han_lp_trainer',
        },
        'RoHe': 'RoHe_trainer',
        'Mg2vec': 'mg2vec_trainer',
        'DHNE': 'DHNE_trainer',
        'DiffMG': 'DiffMG_trainer',
        'MeiREC': 'MeiREC_trainer',
        'KGAT': 'KGAT_trainer'
        'SHGP': 'SHGP_trainer'
        'HGCL': 'hgcltrainer',
    }
    immutable_params = ['model', 'dataset', 'task']

    def __init__(self, model, dataset, task,
                 gpu: int = -1,
                 use_best_config: bool = False,
                 load_from_pretrained: bool = False,
                 hpo_search_space=None,
                 hpo_trials: int = 100,
                 output_dir: str = "./openhgnn/output",
                 conf_path: str = default_conf_path,
                 **kwargs):
        self.config = Config(file_path=conf_path, model=model, dataset=dataset, task=task, gpu=gpu)
        self.config.model = model
        self.config.dataset = dataset
        self.config.task = task
        self.config.gpu = gpu
        self.config.use_best_config = use_best_config
        # self.config.use_hpo = use_hpo
        self.config.load_from_pretrained = load_from_pretrained
        self.config.output_dir = os.path.join(output_dir, self.config.model_name)
        # self.config.seed = seed
        self.config.hpo_search_space = hpo_search_space
        self.config.hpo_trials = hpo_trials

        if not getattr(self.config, 'seed', False):
            self.config.seed = 0
        if use_best_config:
            self.config = set_best_config(self.config)
        self.set_params(**kwargs)
        print(self)

    def set_params(self, **kwargs):
        for key, value in kwargs.items():
            assert key not in self.immutable_params
            self.config.__setattr__(key, value)

    def run(self):
        """ run the experiment """

        # 'line_profiler_func' is for internal use in profiling code execution time, here is an example to use it:
        # from openhgnn import Experiment
        # from openhgnn.trainerflow import NodeClassification
        # Experiment(model='RGCN', dataset='acm4GTN', task='node_classification', gpu=-1, max_epoch=1, line_profiler_func=[NodeClassification.train, ]).run()

        if hasattr(self.config, 'line_profiler_func'):
            from line_profiler import LineProfiler
            prof = LineProfiler
            for func in self.config.line_profiler_func:
                prof = prof(func)
            prof.enable_by_count()

        self.config.logger = Logger(self.config)
        set_random_seed(self.config.seed)
        trainerflow = self.specific_trainerflow.get(self.config.model, self.config.task)
        if type(trainerflow) is not str:
            trainerflow = trainerflow.get(self.config.task)
        if self.config.hpo_search_space is not None:
            # hyper-parameter search
            hpo_experiment(self.config, trainerflow)
        else:
            flow = build_flow(self.config, trainerflow)
            result = flow.train()
            if hasattr(self.config, 'line_profiler_func'):
                prof.print_stats()
            return result

    def __repr__(self):
        basic_info = '------------------------------------------------------------------------------\n' \
                     ' Basic setup of this experiment: \n' \
                     '     model: {}    \n' \
                     '     dataset: {}   \n' \
                     '     task: {}. \n' \
                     ' This experiment has following parameters. You can use set_params to edit them.\n' \
                     ' Use print(experiment) to print this information again.\n' \
                     '------------------------------------------------------------------------------\n'. \
            format(self.config.model_name, self.config.dataset_name, self.config.task)
        params_info = ''
        for attr in dir(self.config):
            if '__' not in attr and attr not in self.immutable_params:
                params_info += '{}: {}\n'.format(attr, getattr(self.config, attr))
        return basic_info + params_info



import torch.nn as nn
import torch.nn.functional as F
import dgl
from dgl.nn.pytorch import GATConv
from . import BaseModel, register_model
from ..layers.macro_layer.SemanticConv import SemanticAttention
from ..layers.MetapathConv import MetapathConv
from ..utils.utils import extract_metapaths, get_ntypes_from_canonical_etypes


@register_model('HAN')
class HAN(BaseModel):
    r"""
    This model shows an example of using dgl.metapath_reachable_graph on the original heterogeneous
    graph HAN from paper `Heterogeneous Graph Attention Network <https://arxiv.org/pdf/1903.07293.pdf>`__..
    Because the original HAN implementation only gives the preprocessed homogeneous graph, this model
    could not reproduce the result in HAN as they did not provide the preprocessing code, and we
    constructed another dataset from ACM with a different set of papers, connections, features and
    labels.


    .. math::
        \mathbf{h}_{i}^{\prime}=\mathbf{M}_{\phi_{i}} \cdot \mathbf{h}_{i}

    where :math:`h_i` and :math:`h'_i` are the original and projected feature of node :math:`i`

    .. math::
        e_{i j}^{\Phi}=a t t_{\text {node }}\left(\mathbf{h}_{i}^{\prime}, \mathbf{h}_{j}^{\prime} ; \Phi\right)

    where :math:`{att}_{node}` denotes the deep neural network.

    .. math::
        \alpha_{i j}^{\Phi}=\operatorname{softmax}_{j}\left(e_{i j}^{\Phi}\right)=\frac{\exp \left(\sigma\left(\mathbf{a}_{\Phi}^{\mathrm{T}} \cdot\left[\mathbf{h}_{i}^{\prime} \| \mathbf{h}_{j}^{\prime}\right]\right)\right)}{\sum_{k \in \mathcal{N}_{i}^{\Phi}} \exp \left(\sigma\left(\mathbf{a}_{\Phi}^{\mathrm{T}} \cdot\left[\mathbf{h}_{i}^{\prime} \| \mathbf{h}_{k}^{\prime}\right]\right)\right)}

    where :math:`\sigma` denotes the activation function, || denotes the concatenate
    operation and :math:`a_{\Phi}` is the node-level attention vector for meta-path :math:`\Phi`.

    .. math::
        \mathbf{z}_{i}^{\Phi}=\prod_{k=1}^{K} \sigma\left(\sum_{j \in \mathcal{N}_{i}^{\Phi}} \alpha_{i j}^{\Phi} \cdot \mathbf{h}_{j}^{\prime}\right)

    where :math:`z^{\Phi}_i` is the learned embedding of node i for the meta-path :math:`\Phi`.
    Given the meta-path set {:math:`\Phi_0 ,\Phi_1,...,\Phi_P`},after feeding node features into node-level attentionwe can obtain P groups of
    semantic-specific node embeddings, denotes as {:math:`Z_0 ,Z_1,...,Z_P`}.
    We use MetapathConv to finish Node-level Attention and Semantic-level Attention.


    Parameters
    ------------
    ntype_meta_paths_dict : dict[str, dict[str, list[etype]]]
        Dict from node type to dict from meta path name to meta path. For node classification, there is only one node type.
        For link prediction, there can be multiple node types which are source and destination node types of target links.
    in_dim : int
        Input feature dimension.
    hidden_dim : int
        Hidden layer dimension.
    out_dim : int
        Output feature dimension.
    num_heads : list[int]
        Number of attention heads.
    dropout : float
        Dropout probability.
    """

    @classmethod
    def build_model_from_args(cls, args, hg):

        # build ntype_meta_paths_dict

        ntypes = set()
        if hasattr(args, 'target_link'):
            ntypes = get_ntypes_from_canonical_etypes(args.target_link)
        elif hasattr(args, 'category'):
            ntypes.add(args.category)
        else:
            raise ValueError

        ntype_meta_paths_dict = {}
        for ntype in ntypes:
            ntype_meta_paths_dict[ntype] = {}
            for meta_path_name, meta_path in args.meta_paths_dict.items():
                # a meta path starts with this node type
                if meta_path[0][0] == ntype:
                    ntype_meta_paths_dict[ntype][meta_path_name] = meta_path
        for ntype, meta_paths_dict in ntype_meta_paths_dict.items():
            if len(meta_paths_dict) == 0:
                ntype_meta_paths_dict[ntype] = extract_metapaths(ntype, hg.canonical_etypes)

        return cls(ntype_meta_paths_dict=ntype_meta_paths_dict,
                   in_dim=args.hidden_dim,
                   hidden_dim=args.hidden_dim,
                   out_dim=args.out_dim,
                   num_heads=args.num_heads,
                   dropout=args.dropout)

    def __init__(self, ntype_meta_paths_dict, in_dim, hidden_dim, out_dim, num_heads, dropout):
        super(HAN, self).__init__()
        self.out_dim = out_dim
        self.mod_dict = nn.ModuleDict()
        for ntype, meta_paths_dict in ntype_meta_paths_dict.items():
            self.mod_dict[ntype] = _HAN(meta_paths_dict, in_dim, hidden_dim, out_dim, num_heads, dropout)

    def forward(self, g, h_dict):
        r"""
        Parameters
        -----------
        g : DGLHeteroGraph or dict[str, dict[str, DGLBlock]]
            For full batch, it is a heterogeneous graph. For mini batch, it is a dict from node type to dict from
            mata path name to DGLBlock.
        h_dict : dict[str, Tensor] or dict[str, dict[str, dict[str, Tensor]]]
            The input features. For full batch, it is a dict from node type to node features. For mini batch, it is
            a dict from node type to dict from meta path name to dict from node type to node features.

        Returns
        --------
        out_dict : dict[str, Tensor]
            The output features. Dict from node type to node features.
        """
        out_dict = {}
        for ntype, han in self.mod_dict.items():
            if isinstance(g, dict):
                # mini batch
                if ntype not in g:
                    continue
                _g = g[ntype]
                _in_h = h_dict[ntype]
            else:
                # full batch
                _g = g
                _in_h = h_dict
            _out_h = han(_g, _in_h)
            for ntype, h in _out_h.items():
                out_dict[ntype] = h

        return out_dict


class _HAN(nn.Module):

    def __init__(self, meta_paths_dict, in_dim, hidden_dim, out_dim, num_heads, dropout):
        super(_HAN, self).__init__()
        self.layers = nn.ModuleList()
        self.layers.append(HANLayer(meta_paths_dict, in_dim, hidden_dim, num_heads[0], dropout))
        for l in range(1, len(num_heads)):
            self.layers.append(HANLayer(meta_paths_dict, hidden_dim * num_heads[l - 1],
                                        hidden_dim, num_heads[l], dropout))
        self.linear = nn.Linear(hidden_dim * num_heads[-1], out_dim)

    def forward(self, g, h_dict):
        for gnn in self.layers:
            h_dict = gnn(g, h_dict)
        out_dict = {}
        for ntype, h in h_dict.items():  # only one ntype here
            out_dict[ntype] = self.linear(h_dict[ntype])
        return out_dict

    def get_emb(self, g, h_dict):
        h = h_dict[self.category]
        for gnn in self.layers:
            h = gnn(g, h)

        return {self.category: h.detach().cpu().numpy()}


class HANLayer(nn.Module):
    """
    HAN layer.

    Parameters
    ------------
    meta_paths_dict : dict[str, list[etype]]
        Dict from meta path name to meta path.
    in_dim : int
        Input feature dimension.
    hidden_dim : int
        Hidden layer dimension.
    out_dim : int
        Output feature dimension.
    layer_num_heads : int
        Number of attention heads.
    dropout : float
        Dropout probability.

    Attributes
    ------------
    _cached_graph : dgl.DGLHeteroGraph
        a cached graph
    _cached_coalesced_graph : list
        _cached_coalesced_graph list generated by *dgl.metapath_reachable_graph()*
    """

    def __init__(self, meta_paths_dict, in_dim, out_dim, layer_num_heads, dropout):
        super(HANLayer, self).__init__()
        self.meta_paths_dict = meta_paths_dict
        semantic_attention = SemanticAttention(in_size=out_dim * layer_num_heads)
        mods = nn.ModuleDict({mp: GATConv(in_dim, out_dim, layer_num_heads,
                                          dropout, dropout, activation=F.elu,
                                          allow_zero_in_degree=True) for mp in meta_paths_dict})
        self.model = MetapathConv(meta_paths_dict, mods, semantic_attention)
        self._cached_graph = None
        self._cached_coalesced_graph = {}

    def forward(self, g, h):
        r"""
        Parameters
        -----------
        g : DGLHeteroGraph or dict[str, DGLBlock]
            For full batch, it is a heterogeneous graph. For mini batch, it is a dict from mata path name to DGLBlock.
        h : dict[str, Tensor] or dict[str, dict[str, Tensor]]
            The input features. For full batch, it is a dict from node type to node features. For mini batch, it is
            a  dict from meta path name to dict from node type to node features.

        Returns
        --------
        h : dict[str, Tensor]
            The output features. Dict from node type to node features. Only one node type.
        """
        # mini batch
        if isinstance(g, dict):
            h = self.model(g, h)

        # full batch
        else:
            if self._cached_graph is None or self._cached_graph is not g:
                self._cached_graph = g
                self._cached_coalesced_graph.clear()
                for mp, mp_value in self.meta_paths_dict.items():
                    self._cached_coalesced_graph[mp] = dgl.metapath_reachable_graph(
                        g, mp_value)
            h = self.model(self._cached_coalesced_graph, h)

        return h



import dgl
import torch as th
import math
import random


class TransX_Sampler():
    def __init__(self, hg, args):
        random.seed(None)

        self.hg = hg.cpu()
        self.src, self.rel, self.dst = self._hg2triplets(self.hg)
        self.num_edges = self.src.shape[0]
        self.num_nodes = self.hg.num_nodes()

        self.batch_size = args.batch_size
        self.neg_size = args.neg_size
        self.batch_num = math.ceil(self.num_edges / self.batch_size)

        # for filter
        self.filter_set = self._build_filter_set(self.src, self.rel, self.dst)
        self.src, self.rel, self.dst = self._random_select_triplets(self.src, self.rel, self.dst,
                                                                    self.num_edges)  # tensor

        self.cur_idx = 0

    def _hg2triplets(self, hg):
        '''
        Attributes
        -------------
        input:
            hg : dgl.heterograph
                One Node type in hg, various edge types
        output:
            src, rel, dst : th.tensor, th.tensor, th.tensor
                #node, #edge, #node
        '''
        src, rel, dst = [], [], []
        if len(hg.ntypes) > 1 and len(hg.etypes) > 1:
            g = dgl.to_homogeneous(hg)
            src, dst = g.edges()
            rel = g.edata[dgl.ETYPE]
            pass

        else:
            for edge_type in hg.etypes:
                src_nx, dst_nx = hg.all_edges(etype=edge_type)
                rel_nx = th.full_like(src_nx, int(edge_type))
                src.append(src_nx)
                rel.append(rel_nx)
                dst.append(dst_nx)
            src = th.cat(src, dim=-1)
            rel = th.cat(rel, dim=-1)
            dst = th.cat(dst, dim=-1)
        return src, rel, dst

    def _random_select_triplets(self, src, rel, dst, num):
        l = list(range(len(src)))
        idx = random.sample(l, num)
        return src[idx], rel[idx], dst[idx]

    def get_pos(self):
        if self.cur_idx + self.batch_size < self.num_edges:
            edge_idx = th.arange(self.cur_idx, self.cur_idx + self.batch_size, dtype=th.int64)
            shuffle = False
        else:
            edge_idx = th.arange(self.cur_idx, self.num_edges, dtype=th.int64)
            shuffle = True
        self.cur_idx = (self.cur_idx + self.batch_size) % self.num_edges
        self.pos_g = (self.src[edge_idx], self.rel[edge_idx], self.dst[edge_idx])
        if shuffle:
            self._random_select_triplets(self.src, self.rel, self.dst, self.num_edges)
            edge_idx = 0
        return self.pos_g

    def get_neg(self):
        srcs, rels, dsts = self.pos_g
        srcs, rels, dsts = srcs.tolist(), rels.tolist(), dsts.tolist()
        node_list = list(range(self.num_nodes))
        replace_node = random.sample(node_list, self.neg_size)

        new_rel = [x for x in rels for i in range(self.neg_size)]
        new_src = [x for x in srcs for i in range(self.neg_size)]
        new_dst = [x for x in dsts for i in range(self.neg_size)]

        batch_size = len(rels)
        part_size = batch_size * self.neg_size

        rep_h = replace_node * batch_size
        for i in range(part_size):
            src, rel, dst = rep_h[i], new_rel[i], new_dst[i]
            while (src, rel, dst) in self.filter_set:
                rep_h[i] = src = random.randint(0, self.num_nodes - 1)

        rep_t = replace_node * batch_size
        for i in range(part_size):
            src, rel, dst = new_src[i], new_rel[i], rep_t[i]
            while (src, rel, dst) in self.filter_set:
                rep_t[i] = dst = random.randint(0, self.num_nodes - 1)

        new_src = rep_h + new_src
        new_rel = new_rel * 2
        new_dst = new_dst + rep_t

        return th.tensor(new_src), th.tensor(new_rel), th.tensor(new_dst)

    def _build_filter_set(self, srcs, rels, dsts):
        srcs, rels, dsts = srcs.tolist(), rels.tolist(), dsts.tolist()
        st = set()
        for i in range(len(srcs)):
            src, rel, dst = srcs[i], rels[i], dsts[i]
            st.add((src, rel, dst))
        return st



import torch
import torch as th
import os
from dgl.data import DGLBuiltinDataset
from dgl.data.utils import load_graphs, save_graphs
import csv
from tqdm import tqdm
import numpy as np
import pickle as pkl
import dgl

__all__ = ['AliRCDDataset', 'AliRCDSmallDataset', 'AliRCDSession1Dataset', 'AliRCDSession2Dataset','AliICDMDataset']


class AliRCDDataset(DGLBuiltinDataset):
    r"""AliRCD(Alibaba Risk Commodity Detection) dataset is extracted from real-world risk scenarios at Alibaba.
    This dataset is for ICDM 2022 competition. Detailed Information is here:
    https://tianchi.aliyun.com/competition/entrance/531976/information. When we create the dgl graph, we rearrange
    the node ids from the original files. The map from original node id to dgl node id can be retrieved from property
    item_map and the map from dgl node id to original node id can be retrieved from property rev_item_map.

    Parameters
    ----------
    session : str
        'small', 'session1' or 'session2'. The small one is only used for debug and helps with understanding
        the data format.
    raw_dir : str
        Specifying the directory that will store the
        downloaded data or the directory that
        already stores the input data.
        Default: ~/.dgl/
    force_reload : bool
        Whether to reload the dataset. Default: False
    verbose : bool
        Whether to print out progress information. Default: False
    transform : callable, optional
        A transform that takes in a :class:`~dgl.DGLGraph` object and returns
        a transformed version. The :class:`~dgl.DGLGraph` object will be
        transformed before every access.
    """

    def __init__(self, session, raw_dir=None, force_reload=False,
                 verbose=False, transform=None):
        name = 'AliRCD_{}'.format(session)
        self.session = session
        self.load_labels = not session == 'session2'

        super(AliRCDDataset, self).__init__(
            name,
            url='https://s3.cn-north-1.amazonaws.com.cn/dgl-data/dataset/openhgnn/{}.zip'.format(name),
            raw_dir=raw_dir,
            force_reload=force_reload, verbose=verbose, transform=transform)

    def process(self):
        if self.session == 'session1':
            edge_size = 157814864
            node_size = 13806619
        elif self.session == 'session2':
            edge_size = 120691444
            node_size = 10284026
        else:  # for debug
            edge_size = 100
            node_size = 100

        # load node info
        nodes_info = self._get_node_atts(node_size)

        # item map and reversed item map
        self._item_map = nodes_info['maps']['item']
        self._rev_item_map = {}
        for k, v in self._item_map.items():
            self._rev_item_map[v] = k

        # load edges
        g = self._format_dgl_edges(nodes_info, edge_size)

        # add attrs
        for ntype, embedding_dict in nodes_info['embeds'].items():
            dim = embedding_dict[0].shape[0]
            g.nodes[ntype].data['h'] = torch.rand(g.num_nodes(ntype), dim)
            for nid, embedding in tqdm(embedding_dict.items()):
                g.nodes[ntype].data['h'][nid] = torch.from_numpy(embedding)
        # load label
        num_nodes = g.num_nodes(self.category)
        if self.load_labels:
            labels_path = os.path.join(self.save_path, '{}_train_labels.csv'.format(self.name))
            labels = th.tensor([float('nan')] * g.num_nodes(self.category))
            with open(labels_path, 'r') as f:
                csvreader = csv.reader(f)
                for row in csvreader:
                    orig_id = int(row[0])
                    new_id = self._item_map.get(orig_id)
                    if new_id is not None:
                        labels[new_id] = int(row[1])
            label_mask = ~th.isnan(labels)
            label_idx = th.nonzero(label_mask, as_tuple=False).squeeze()
            g.nodes[self.category].data['label'] = labels.type(th.int64)

            # label_idx = np.random.permutation(np.array(label_idx))  # shuffle the label index
            split_ratio = [0.8, 0.2]
            num_labels = len(label_idx)
            train_mask = th.zeros(num_nodes).bool()
            train_mask[label_idx[0: int(split_ratio[0] * num_labels)]] = True
            val_mask = th.zeros(num_nodes).bool()
            val_mask[
                label_idx[int(split_ratio[0] * num_labels): int((split_ratio[0] + split_ratio[1]) * num_labels)]] = True
            g.nodes[self.category].data['train_mask'] = train_mask
            g.nodes[self.category].data['val_mask'] = val_mask

        # load test_idx
        test_idx_path = os.path.join(self.save_path, '{}_test_ids.csv'.format(self.name))
        test_mask = th.zeros(num_nodes).bool()
        with open(test_idx_path, 'r') as f:
            csvreader = csv.reader(f)
            for row in csvreader:
                orig_id = int(row[0])
                new_id = self._item_map.get(orig_id)
                if new_id is not None:
                    test_mask[new_id] = True
        g.nodes[self.category].data['test_mask'] = test_mask

        if self.session == "ICDM" :
            test_labels_path = os.path.join(self.save_path, '{}_test_labels.csv'.format(self.name))
            with open(test_labels_path, 'r') as f:
                csvreader = csv.reader(f)
                for row in csvreader:
                    line = row[0].split('\t')
                    test_label = int(float(line[1]))
                    test_id = int(line[0])
                    new_id = self._item_map.get(test_id)
                    if new_id is not None:
                        g.nodes[self.category].data['label'][new_id] = test_label

        self._g = g
        if self.verbose:
            print(self._g)
            print('finish loading dataset')

    def _get_node_atts(self, node_size=100):
        node_file = os.path.join(self.save_path, '{}_nodes.csv'.format(self.name))
        node_maps = {}
        node_embeds = {}
        count = 0
        count2 = 0
        node_counts = node_size
        process = tqdm(total=node_counts)
        with open(node_file, 'r') as rf:
            while True:
                line = rf.readline()
                if line is None or len(line) == 0:
                    break
                info = line.strip().split(",")

                node_id = int(info[0])
                node_type = info[1].strip()
                node_maps.setdefault(node_type, {})
                node_id_v2 = len(node_maps[node_type])
                node_maps[node_type][node_id] = node_id_v2
                if node_type == 'item' and len(info[2]) == 0:
                    node_embeds.setdefault(node_type, {})
                    node_embeds[node_type][node_id_v2] = np.zeros(128, dtype=np.float32)
                    count2 += 1
                elif len(info) == 3 and len(info[2]) > 0:
                    node_embeds.setdefault(node_type, {})
                    node_embeds[node_type][node_id_v2] = np.array([x for x in info[2].split(":")], dtype=np.float32)
                count += 1
                if count % 100000 == 0:
                    process.update(100000)
        process.close()

        print('lack of features:', count2, count, len(node_maps['item']))
        print('node_types', node_maps.keys())
        for node_type in node_maps:
            print('node_type', node_type, len(node_maps[node_type]))
        print(len(node_embeds['item']))

        nodes_dict = {'maps': node_maps, 'embeds': node_embeds}
        return nodes_dict

    def _format_dgl_edges(self, node_info, edge_size=100):
        node_maps = node_info['maps']
        edges = {}
        edges_hg = {}

        process = tqdm(total=edge_size)
        count = 0
        edge_file = os.path.join(self.save_path, '{}_edges.csv'.format(self.name))
        with open(edge_file, 'r') as rf:
            while True:
                line = rf.readline()
                if line is None or len(line) == 0:
                    break
                line_info = line.strip().split(",")
                source_id, dest_id, source_type, dest_type, edge_type = line_info
                source_id = node_maps[source_type][int(source_id)]
                dest_id = node_maps[dest_type][int(dest_id)]
                edges.setdefault(edge_type, {})
                edges[edge_type].setdefault('source', []).append(source_id)
                edges[edge_type].setdefault('dest', []).append(dest_id)
                edges[edge_type].setdefault('source_type', source_type)
                edges[edge_type].setdefault('dest_type', dest_type)
                count += 1
                if count % 100000 == 0:
                    process.update(100000)
        process.close()

        for edge_type in edges:
            source_type = edges[edge_type]['source_type']
            dest_type = edges[edge_type]['dest_type']
            source = edges[edge_type]['source']
            dest = edges[edge_type]['dest']
            edges_hg[(source_type, edge_type, dest_type)] = list(zip(source, dest))

        hg = dgl.heterograph(edges_hg)
        return hg

    def has_cache(self):
        graph_path = os.path.join(self.save_path, 'graph.bin')
        return os.path.exists(graph_path)

    def save(self):
        graph_path = os.path.join(self.save_path, 'graph.bin')
        save_graphs(graph_path, self._g)
        pkl.dump(self.item_map, open(os.path.join(self.save_path, 'item_map.pkl'), 'wb'))
        pkl.dump(self.rev_item_map, open(os.path.join(self.save_path, 'rev_item_map.pkl'), 'wb'))

    def load(self):
        graph_path = os.path.join(self.save_path, 'graph.bin')
        gs, _ = load_graphs(graph_path)
        self._item_map = pkl.load(open(os.path.join(self.save_path, 'item_map.pkl'), 'rb'))
        self._rev_item_map = pkl.load(open(os.path.join(self.save_path, 'rev_item_map.pkl'), 'rb'))
        self._g = gs[0]
        if self.verbose:
            print(self._g)

    @property
    def item_map(self):
        return self._item_map

    @property
    def rev_item_map(self):
        return self._rev_item_map

    @property
    def category(self):
        return 'item'

    @property
    def target_ntype(self):
        return 'item'

    @property
    def num_classes(self):
        return 2

    def __getitem__(self, idx):
        assert idx == 0
        return self._g

    def __len__(self):
        return 1

    @property
    def meta_paths_dict(self):
        return {'IBI': [('item', 'A', 'b'),
                        ('b', 'A_1', 'item')],
                'IFAFI': [('item', 'B_1', 'f'),
                          ('f', 'G', 'a'),
                          ('a', 'G_1', 'f'),
                          ('f', 'B', 'item')],
                'IFCFI': [('item', 'B_1', 'f'),
                          ('f', 'D', 'c'),
                          ('c', 'D_1', 'f'),
                          ('f', 'B', 'item')],
                'IFDFI': [('item', 'B_1', 'f'),
                          ('f', 'C', 'd'),
                          ('d', 'C_1', 'f'),
                          ('f', 'B', 'item')],
                'IFEFI': [('item', 'B_1', 'f'),
                          ('f', 'F', 'e'),
                          ('e', 'F_1', 'f'),
                          ('f', 'B', 'item')],
                }


class AliRCDSmallDataset(AliRCDDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        session = 'small'
        super(AliRCDSmallDataset, self).__init__(session, raw_dir=raw_dir, force_reload=force_reload, verbose=verbose,
                                                 transform=transform)


class AliRCDSession1Dataset(AliRCDDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        session = 'session1'
        super(AliRCDSession1Dataset, self).__init__(session, raw_dir=raw_dir, force_reload=force_reload,
                                                    verbose=verbose,
                                                    transform=transform)


class AliRCDSession2Dataset(AliRCDDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        session = 'session2'
        super(AliRCDSession2Dataset, self).__init__(session, raw_dir=raw_dir, force_reload=force_reload,
                                                    verbose=verbose,
                                                    transform=transform)
        
class AliICDMDataset(AliRCDDataset):
    def __init__(self, raw_dir=None, force_reload=False, verbose=False, transform=None):
        session = 'ICDM'
        super(AliICDMDataset, self).__init__(session, raw_dir=raw_dir, force_reload=force_reload,
                                                    verbose=verbose,
                                                    transform=transform)


import os
import dgl
import torch as th
from . import BaseDataset, register_dataset
from dgl.data.utils import download,load_graphs,save_graphs,save_info,load_info
from .multigraph import MultiGraphDataset
from ..sampler.negative_sampler import Uniform_exclusive
from . import AcademicDataset
from .HGCLDataset import HGCLDataset

#add more lib for KGAT
import time
import collections
import zipfile
import numpy as np
import pandas as pd

@register_dataset('recommendation')
class RecommendationDataset(BaseDataset):
    """

    """
    def __init__(self,*args, **kwargs):
        super(RecommendationDataset, self).__init__(*args, **kwargs)
        self.meta_paths_dict = None


@register_dataset('kgcn_recommendation')
class KGCN_Recommendation(RecommendationDataset):
    r"""
    Which is used in KGCN.
    """
    def __init__(self, dataset_name, *args, **kwargs):
            super(RecommendationDataset, self).__init__(*args, **kwargs)
            dataset = MultiGraphDataset(name=dataset_name, raw_dir='')
            self.g = dataset[0].long()
            self.g_1 = dataset[1].long()

    def get_split(self, validation=True):
        ratingsGraph = self.g_1
        n_edges = ratingsGraph.num_edges()
        random_int = th.randperm(n_edges)
        train_idx = random_int[:int(n_edges*0.6)]
        val_idx = random_int[int(n_edges*0.6):int(n_edges*0.8)]
        test_idx = random_int[int(n_edges*0.6):int(n_edges*0.8)]

        return train_idx, val_idx, test_idx
    
    def get_train_data(self):
        pass

    def get_labels(self):
        return self.label

@register_dataset('hgcl_recommendation')
class HGCLRecommendation(RecommendationDataset):
    def __init__(self, dataset_name, *args, **kwargs):
        super(RecommendationDataset, self).__init__(*args, **kwargs)
        dataset = HGCLDataset(name=dataset_name, raw_dir='')
        self.g = dataset[0].long()

    def get_split(self, validation=True):
        ratingsGraph = self.g
        n_edges = ratingsGraph.num_edges()
        random_int = th.randperm(n_edges)
        train_idx = random_int[:int(n_edges * 0.6)]
        val_idx = random_int[int(n_edges * 0.6):int(n_edges * 0.8)]
        test_idx = random_int[int(n_edges * 0.6):int(n_edges * 0.8)]

        return train_idx, val_idx, test_idx

    def get_train_data(self):
        pass

    def get_labels(self):
        return self.label
        
@register_dataset('hin_recommendation')
class HINRecommendation(RecommendationDataset):
    def __init__(self, dataset_name, *args, **kwargs):
        super(HINRecommendation, self).__init__(*args, **kwargs)
        self.dataset_name = dataset_name
        self.num_neg = 20
        #self.neg_dir = os.path.join(self.raw_dir, dataset_name, 'neg_{}.bin'.format(self.num_neg))
        if dataset_name == 'yelp4rec':
            dataset = AcademicDataset(name='yelp4rec', raw_dir='')
            self.g = dataset[0].long()
            self.target_link = 'user-item'
            self.target_link_r = 'item-user'
            self.user_name = 'user'
            self.item_name = 'item'

        elif dataset_name == 'yelp4HeGAN':
            dataset = AcademicDataset(name='yelp4HeGAN', raw_dir='')
            self.g = dataset[0].long()
            self.target_link = 'usb'
            self.target_link_r = 'bus'
            self.user_name = 'user'
            self.item_name = 'business'

        elif dataset_name == 'DoubanMovie':
            dataset = AcademicDataset(name='DoubanMovie', raw_dir='')
            self.g = dataset[0].long()
            self.target_link = 'user-item'
            self.target_link_r = 'item-user'
            self.user_name = 'user'
            self.item_name = 'item'

        elif dataset_name == 'amazon':
            data_path = './openhgnn/dataset/amazon_graph.bin'
            dataset = AcademicDataset(name='amazon', raw_dir='')
            self.g = dataset[0].long()
            self.target_link = 'ui'
            self.target_link_r = 'iu'
            self.user_name = 'user'
            self.item_name = 'item'

        self.out_ntypes = [self.user_name, self.item_name]
        # self.process()
        # self.neg_g = self.construct_negative_graph(self.g)

    def load_HIN(self, dataset_name):
        g, _ = dgl.load_graphs(dataset_name)
        return g[0]

    # def process(self, g):
    #     # sub 1 for every node
    #     new = {}
    #     for etype in g.canonical_etypes:
    #         edges = g.edges(etype=etype)
    #         new[etype] = (edges[0]-1, edges[1]-1)
    #     hg = dgl.heterograph(new)
    #     hg.edata['val_mask'] = g.edata['val_mask']
    #     hg.edata['test_mask'] = g.edata['test_mask']
    #     hg.edata['train_mask'] = g.edata['train_mask']
    #     from dgl.data.utils import save_graphs
    #     save_graphs(f"./openhgnn/dataset/{self.dataset_name}.bin", hg)

    def get_split(self, validation=True):
        test_mask = self.g.edges[self.target_link].data['test_mask'].squeeze()
        test_index = th.nonzero(test_mask).squeeze()
        test_edge = self.g.find_edges(test_index, self.target_link)

        test_graph = dgl.heterograph({(self.user_name, self.target_link, self.item_name): test_edge},
                                     {ntype: self.g.number_of_nodes(ntype) for ntype in self.out_ntypes})
        if validation:
            val_mask = self.g.edges[self.target_link].data['val_mask'].squeeze()
            val_index = th.nonzero(val_mask).squeeze()
            val_edge = self.g.find_edges(val_index, self.target_link)

            val_graph = dgl.heterograph({(self.user_name, self.target_link, self.item_name): val_edge},
                                             {ntype: self.g.number_of_nodes(ntype) for ntype in self.out_ntypes})

            train_graph = dgl.remove_edges(self.g, th.cat((val_index, test_index)), self.target_link)
            train_graph = dgl.remove_edges(train_graph, th.cat((val_index, test_index)), self.target_link_r)
        else:
            train_graph = dgl.remove_edges(self.g, test_index, self.target_link)
            train_graph = dgl.remove_edges(train_graph, test_index, self.target_link_r)
            val_graph = train_graph

        return train_graph, val_graph, test_graph

    def construct_negative_graph(self, train_g):
        fname = f'./openhgnn/dataset/{self.dataset_name}/neg_graph_{self.num_neg}.bin'
        if os.path.exists(fname):
            g, _ = load_graphs(fname)
            return g[0]
        else:
            k = self.num_neg
            negative_sampler = Uniform_exclusive(k)
            negative_edges = negative_sampler(train_g.to('cpu'), {
                self.target_link: th.arange(train_g.num_edges(self.target_link))})
            # negative_edges = negative_sampler(train_g.to('cpu'), {
            #     self.target_link: th.arange(10)})
            neg_g = dgl.heterograph(negative_edges,
                                    {ntype: self.g.number_of_nodes(ntype) for ntype in self.out_ntypes})
            dgl.save_graphs(fname, neg_g)
            return neg_g


@register_dataset('test_link_prediction')
class Test_Recommendation(RecommendationDataset):
    def __init__(self, dataset_name):
        super(RecommendationDataset, self).__init__()
        self.g = self.load_HIN('./openhgnn/debug/data.bin')
        self.target_link = 'user-item'
        self.has_feature = False
        self.preprocess()
        #self.generate_negative()

    def load_HIN(self, dataset_name):
        g, _ = load_graphs(dataset_name)
        return g[0]

    def preprocess(self):
        test_mask = self.g.edges[self.target_link].data['test_mask']
        index = th.nonzero(test_mask).squeeze()
        self.test_edge = self.g.find_edges(index, self.target_link)
        self.pos_test_graph = dgl.heterograph({('user', 'user-item', 'item'): self.test_edge}, {ntype: self.g.number_of_nodes(ntype) for ntype in ['user', 'item']})
        self.g.remove_edges(index, self.target_link)
        self.g.remove_edges(index, 'item-user')
        self.neg_test_graph, _ = dgl.load_graphs('./openhgnn/debug/neg.bin')
        self.neg_test_graph = self.neg_test_graph[0]
        return

        negative_sampler = Uniform_exclusive(99)
        self.negative_g = negative_sampler(self.hg.to('cpu'), {self.target_link: th.arange(self.hg.num_edges(self.target_link))})
    def generate_negative(self):
        k = 99
        e = self.pos_test_graph.edges()
        neg_src = []
        neg_dst = []
        for i in range(self.pos_test_graph.number_of_edges()):
            src = e[0][i]
            exp = self.pos_test_graph.successors(src)
            dst = th.randint(high=self.g.number_of_nodes('item'), size=(k,))
            for d in range(len(dst)):
                while dst[d] in exp:
                    dst[d] = th.randint(high=self.g.number_of_nodes('item'), size=(1,))
            src = src.repeat_interleave(k)
            neg_src.append(src)
            neg_dst.append(dst)
        neg_edge = (th.cat(neg_src), th.cat(neg_dst))
        neg_graph = dgl.heterograph({('user', 'user-item', 'item'): neg_edge}, {ntype: self.g.number_of_nodes(ntype) for ntype in ['user', 'item']})
        dgl.save_graphs('./openhgnn/debug/neg.bin', neg_graph)

@register_dataset('kgat_recommendation')
class KGAT_recommendation(RecommendationDataset):

    def __init__(self,dataset_name,*args, **kwargs):
        super(KGAT_recommendation, self).__init__(*args, **kwargs)
        if dataset_name not in ['yelp2018','amazon-book','last-fm']:
            raise KeyError('Dataset {} is not supported!'.format(dataset_name))
        self.dataset_name=dataset_name

        self.data_path=f'openhgnn/dataset/{self.dataset_name}'
        if self.dataset_name=='yelp2018':
            self.processed_data_path = 'openhgnn/dataset/KGAT_yelp2018'
        elif self.dataset_name=='amazon-book':
            self.processed_data_path = 'openhgnn/dataset/KGAT_amazon-book'
        else:
            self.processed_data_path = 'openhgnn/dataset/KGAT_last-fm'

        if os.path.exists(f"{self.processed_data_path}/graph.bin"):
            if os.path.exists(f"{self.processed_data_path}/other_info.pkl"):
                self.load()
        else:
            self.download()
            self.preprocess()
            self.save()

    def download(self):
        prefix = 'https://raw.githubusercontent.com/xiangwang1223/knowledge_graph_attention_network/master/Data'
        if self.dataset_name =='last-fm':
            required_file = ['train.txt', 'test.txt', 'kg_final.txt']
        else:
            required_file = ['train.txt', 'test.txt', 'kg_final.txt.zip']

        for filename in required_file:
            url = f"{prefix}/{self.dataset_name}/{filename}"
            file_path = f"{self.data_path}/{filename}"
            if not os.path.exists(file_path):
                try:
                    download(url,file_path)
                    if filename == 'kg_final.txt.zip':
                        zip_file = zipfile.ZipFile(file_path)
                        zip_file.extractall(f"{self.data_path}")
                        zip_file.close()
                except BaseException as e:
                    print("\n",e)
                    print("\nNote!   --- If you want to download the file, vpn is required ---")
                    print("If you don't have a vpn, please download the dataset from here: https://github.com/xiangwang1223/knowledge_graph_attention_network/")
                    print("\nAfter downloading the dataset, you need to store the files in the following path: ")
                    print(f"{os.getcwd()}\openhgnn\dataset\{self.dataset_name}\\train.txt")
                    print(f"{os.getcwd()}\openhgnn\dataset\{self.dataset_name}\\test.txt")
                    print(f"{os.getcwd()}\openhgnn\dataset\{self.dataset_name}\kg_final.txt")
                    print(f"{os.getcwd()}\openhgnn\dataset\{self.dataset_name}\pretrain\mf.npz")
                    exit()

        #load pretrain file
        prefix=("https://raw.githubusercontent.com/xiangwang1223/knowledge_graph_attention_network/master/Model/pretrain")
        pretrain_url=f"{prefix}/{self.dataset_name}/mf.npz"
        self.pretrain_embedding_dir=f"{self.data_path}/pretrain/mf.npz"
        if not os.path.exists(self.pretrain_embedding_dir):
            download(pretrain_url,self.pretrain_embedding_dir)
        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- All files have been downloaded! ---')

    def preprocess(self):

        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- Start Processing! ---')

        # get data_file_path
        train_file = f"{self.data_path}/train.txt"
        test_file = f"{self.data_path}/test.txt"
        kg_file = f"{self.data_path}/kg_final.txt"

        self.cf_train_data, self.train_user_dict = self.load_cf(train_file)
        self.cf_test_data, self.test_user_dict = self.load_cf(test_file)
        self.statistic_cf()
        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- cf data finish ---')

        kg_data = self.load_kg(kg_file)
        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- kg data load ---')

        self.construct_data(kg_data)
        self.g = self.create_graph(self.kg_train_data, self.n_users_entities)
        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- kg data finish ---')
        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- Done! ---')

    def load_cf(self, filename):
        '''
        Create user-item bipartite graph
        '''
        user = []
        item = []
        user_dict = dict()
        lines = open(filename, 'r').readlines()
        for l in lines:
            tmp = l.strip()
            inter = [int(i) for i in tmp.split()]

            if len(inter) > 1:
                user_id, item_ids = inter[0], inter[1:]
                item_ids = list(set(item_ids))
                for item_id in item_ids:
                    user.append(user_id)
                    item.append(item_id)
                user_dict[user_id] = item_ids

        user = np.array(user, dtype=np.int32)
        item = np.array(item, dtype=np.int32)
        return (user, item), user_dict

    def statistic_cf(self):
        '''
        Some statistic information of the user-item bipartite graph
        '''
        self.n_users = max(max(self.cf_train_data[0]), max(self.cf_test_data[0])) + 1
        self.n_items = max(max(self.cf_train_data[1]), max(self.cf_test_data[1])) + 1
        self.n_cf_train = len(self.cf_train_data[0])
        self.n_cf_test = len(self.cf_test_data[0])

    def load_kg(self, filename):
        '''
        Load knowledge graph data
        '''
        kg_data = pd.read_csv(filename, sep=' ', names=['h', 'r', 't'], engine='python')
        kg_data = kg_data.drop_duplicates()
        return kg_data

    def construct_data(self, kg_data):
        '''
        Alignment of knowledge graph and the user-item bipartite graph
        '''
        n_relations = max(kg_data['r']) + 1
        reverse_kg_data = kg_data.copy()
        reverse_kg_data = reverse_kg_data.rename({'h': 't', 't': 'h'}, axis='columns')
        reverse_kg_data['r'] += n_relations
        kg_data = pd.concat([kg_data, reverse_kg_data], axis=0, ignore_index=True, sort=False)
        kg_data['r'] += 2
        self.n_relations = max(kg_data['r']) + 1
        self.n_entities = max(max(kg_data['h']), max(kg_data['t'])) +1
        self.n_users_entities = self.n_users + self.n_entities

        self.cf_train_data = \
            (np.array(list(map(lambda d: d + self.n_entities, self.cf_train_data[0]))).astype(np.int32), self.cf_train_data[1].astype(np.int32))
        self.cf_test_data = \
            (np.array(list(map(lambda d: d + self.n_entities, self.cf_test_data[0]))).astype(np.int32), self.cf_test_data[1].astype(np.int32))

        self.train_user_dict = {k + self.n_entities: np.unique(v).astype(np.int32) for k, v in self.train_user_dict.items()}
        self.test_user_dict = {k + self.n_entities: np.unique(v).astype(np.int32) for k, v in self.test_user_dict.items()}

        cf2kg_train_data = pd.DataFrame(np.zeros((self.n_cf_train, 3), dtype=np.int32), columns=['h', 'r', 't'])
        cf2kg_train_data['h'] = self.cf_train_data[0]
        cf2kg_train_data['t'] = self.cf_train_data[1]

        reverse_cf2kg_train_data = pd.DataFrame(np.ones((self.n_cf_train, 3), dtype=np.int32), columns=['h', 'r', 't'])
        reverse_cf2kg_train_data['h'] = self.cf_train_data[1]
        reverse_cf2kg_train_data['t'] = self.cf_train_data[0]

        self.kg_train_data = pd.concat([kg_data, cf2kg_train_data, reverse_cf2kg_train_data], ignore_index=True)

        self.n_kg_train = len(self.kg_train_data)

        # construct kg dict
        self.train_kg_dict = collections.defaultdict(list)
        for row in self.kg_train_data.iterrows():
            h, r, t = row[1]
            self.train_kg_dict[h].append((t, r))

    def create_graph(self, kg_data, n_nodes):
        '''
        Create DGLgraph
        '''
        g = dgl.graph((kg_data['t'], kg_data['h']))
        g.ndata['id'] = th.arange(n_nodes, dtype=th.long)
        g.edata['type'] = th.LongTensor(kg_data['r'])
        return g

    def get_split(self):
        '''
        Make it compatible to recommendation task
        '''
        return None,None,None

    def get_labels(self):
        '''
        Make it compatible to recommendation task
        '''
        return None

    def save(self):
        '''
        Store all the information for training
        '''
        #save a dictionary
        print(f'storing graph and other information in {self.processed_data_path}')
        info_dict={'n_users':self.n_users,'n_entities':self.n_entities,'n_relations':self.n_relations,'n_items':self.n_items,
                   'train_user_dict':self.train_user_dict,'test_user_dict':self.test_user_dict,'train_kg_dict':self.train_kg_dict,
                   'n_cf_train':self.n_cf_train,'n_kg_train':self.n_kg_train}
        os.makedirs(os.path.dirname(f"{self.processed_data_path}/other_info.pkl"),exist_ok=True)
        save_info(f"{self.processed_data_path}/other_info.pkl",info_dict)
        save_graphs(f"{self.processed_data_path}/graph.bin",[self.g])

    def load(self):
        '''
        Store all the information for training
        '''
        #load information from a dictionary
        print(f'loading data from {self.processed_data_path}')
        new_dict=load_info(f"{self.processed_data_path}/other_info.pkl")

        self.train_user_dict=new_dict['train_user_dict']
        self.test_user_dict=new_dict['test_user_dict']
        self.train_kg_dict = new_dict['train_kg_dict']

        self.n_users=new_dict['n_users']
        self.n_entities=new_dict['n_entities']
        self.n_relations=new_dict['n_relations']
        self.n_items=new_dict['n_items']

        self.n_cf_train=new_dict['n_cf_train']
        self.n_kg_train =new_dict['n_kg_train']

        self.g=load_graphs(f"{self.processed_data_path}/graph.bin")
        self.g=self.g[0][0]

        self.pretrain_embedding_dir = f"{self.data_path}/pretrain/mf.npz"

        print(time.strftime("%Y-%m-%d %H:%M:%S ", time.localtime()), '--- Start Training! ---')


import importlib
from .NEW_model import MLP_follow_model
from .base_model import BaseModel
from torch import nn
import sys

sys.path.append("..")

MODEL_REGISTRY = {}


def register_model(name):
    """
    New models types can be added to cogdl with the :func:`register_model`
    function decorator.

    For example::

        @register_model('gat')
        class GAT(BaseModel):
            (...)

    Args:
        name (str): the name of the models
    """

    def register_model_cls(cls):
        if name in MODEL_REGISTRY:
            raise ValueError("Cannot register duplicate models ({})".format(name))
        if not issubclass(cls, BaseModel):
            raise ValueError("Model ({}: {}) must extend BaseModel".format(name, cls.__name__))
        MODEL_REGISTRY[name] = cls
        cls.model_name = name
        return cls

    return register_model_cls


def try_import_model(model):
    if model not in MODEL_REGISTRY:
        if model in SUPPORTED_MODELS:
            importlib.import_module(SUPPORTED_MODELS[model])
        else:
            print(f"Failed to import {model} models.")
            return False
    return True


def build_model(model):
    if isinstance(model, nn.Module):
        if not hasattr(model, 'build_model_from_args'):
            def build_model_from_args(args, hg):
                return model

            model.build_model_from_args = build_model_from_args
        return model
    if not try_import_model(model):
        exit(1)
    return MODEL_REGISTRY[model]


SUPPORTED_MODELS = {
    "CompGCN": "openhgnn.models.CompGCN",
    "HetGNN": "openhgnn.models.HetGNN",
    'RGCN': 'openhgnn.models.RGCN',
    "RGAT": 'openhgnn.models.RGAT',
    'RSHN': 'openhgnn.models.RSHN',
    'Metapath2vec': 'openhgnn.models.SkipGram',
    'HERec': 'openhgnn.models.SkipGram',
    'HAN': 'openhgnn.models.HAN',
    'RoHe': 'openhgnn.models.RoHe',
    'HeCo': 'openhgnn.models.HeCo',
    'HGT': 'openhgnn.models.HGT',
    'GTN': 'openhgnn.models.GTN_sparse',
    'fastGTN': 'openhgnn.models.fastGTN',
    'MHNF': 'openhgnn.models.MHNF',
    'MAGNN': 'openhgnn.models.MAGNN',
    'HeGAN': 'openhgnn.models.HeGAN',
    'NSHE': 'openhgnn.models.NSHE',
    'NARS': 'openhgnn.models.NARS',
    'RHGNN': 'openhgnn.models.RHGNN',
    'HPN': 'openhgnn.models.HPN',
    'KGCN': 'openhgnn.models.KGCN',
    'SLiCE': 'openhgnn.models.SLiCE',
    'HGSL': 'openhgnn.models.HGSL',
    'GCN': 'space4hgnn.homo_models.GCN',
    'GAT': 'space4hgnn.homo_models.GAT',
    'homo_GNN': 'openhgnn.models.homo_GNN',
    'general_HGNN': 'openhgnn.models.general_HGNN',
    'HDE': 'openhgnn.models.HDE',
    'SimpleHGN': 'openhgnn.models.SimpleHGN',
    'GATNE-T': 'openhgnn.models.GATNE',
    'HetSANN': 'openhgnn.models.HetSANN',
    'HGAT': 'openhgnn.models.HGAT',
    'ieHGCN': 'openhgnn.models.ieHGCN',
    'TransE': 'openhgnn.models.TransE',
    'TransH': 'openhgnn.models.TransH',
    'TransR': 'openhgnn.models.TransR',
    'TransD': 'openhgnn.models.TransD',
    'GIE':'openhgnn.models.GIE',
    'GIN':'openhgnn.models.GIN',
    'Rsage': 'openhgnn.models.Rsage',
    'Mg2vec': 'openhgnn.models.MG2vec',
    'DHNE': 'openhgnn.models.DHNE',
    'DiffMG': 'openhgnn.models.DiffMG',
    'MeiREC': 'openhgnn.models.MeiREC',
    'HGNN_AC': 'openhgnn.models.HGNN_AC',
    'AEHCL': 'openhgnn.models.AEHCL',
    'KGAT': 'openhgnn.models.KGAT',
    'SHGP': 'openhgnn.models.ATT_HGCN',
    'DSSL': 'openhgnn.models.DSSL'
    'HGCL': 'openhgnn.models.HGCL'
}

from .HGCL import HGCL
from .CompGCN import CompGCN
from .HetGNN import HetGNN
from .RGCN import RGCN
from .RGAT import RGAT
from .RSHN import RSHN
from .SkipGram import SkipGram
from .HAN import HAN
from .RoHe import RoHe
from .HeCo import HeCo
from .HGT import HGT
from .GTN_sparse import GTN
from .fastGTN import fastGTN
from .MHNF import MHNF
from .MAGNN import MAGNN
from .HeGAN import HeGAN
from .NSHE import NSHE
from .NARS import NARS
from .RHGNN import RHGNN
from .HPN import HPN
from .KGCN import KGCN
from .SLiCE import SLiCE
from .HGSL import HGSL
from .homo_GNN import homo_GNN
from .general_HGNN import general_HGNN
from .HDE import HDE
from .SimpleHGN import SimpleHGN
from .HetSANN import HetSANN
from .ieHGCN import ieHGCN
from .HGAT import HGAT
from .GATNE import GATNE
from .Rsage import Rsage
from .Mg2vec import Mg2vec
from .DHNE import DHNE
from .DiffMG import DiffMG
from .MeiREC import MeiREC
from .HGNN_AC import HGNN_AC
from .KGAT import KGAT
from .DSSL import DSSL

__all__ = [
    'BaseModel',
    'CompGCN',
    'HetGNN',
    'RGCN',
    'RGAT',
    'RSHN',
    'SkipGram',
    'HAN',
    'HeCo',
    'HGT',
    'GTN',
    'fastGTN',
    'MHNF',
    'MAGNN',
    'HeGAN',
    'NSHE',
    'NARS',
    'RHGNN',
    'HPN',
    'KGCN',
    'SLiCE',
    'HGSL',
    'homo_GNN',
    'general_HGNN',
    'HDE',
    'SimpleHGN',
    'GATNE',
    'Rsage',
    'Mg2vec',
    'DHNE',
    'DiffMG',
    'MeiREC',
<<<<<<< HEAD
<<<<<<< HEAD
    'KGAT'
=======
    'ATT_HGCN'
>>>>>>> 50ad83434d39d84a5dc44464f36564b451e941e7
=======
    'KGAT',
    'DSSL'
>>>>>>> e1d95c140eccfdb60975128ed8168dfc5ca6ec1f
]
classes = __all__





import torch.nn as nn
from openhgnn.models import BaseModel, register_model
import torch


@register_model("DHNE")
class DHNE(BaseModel):
    r"""
        **Title:** Structural Deep Embedding for Hyper-Networks

        **Authors:** Ke Tu, Peng Cui, Xiao Wang, Fei Wang, Wenwu Zhu

        DHNE was introduced in `[paper] <https://arxiv.org/abs/1711.10146>`_
        and parameters are defined as follows:

        Parameters
        ----------
        nums_type : list
            the type of nodes
        dim_features : array
            the embedding dimension of nodes
        embedding_sizes : int
            the embedding dimension size
        hidden_size : int
            The hidden full connected layer size
        device : int
            the device DHNE working on
        """
    @classmethod
    def build_model_from_args(cls, args):
        return cls(dim_features=args.dim_features,
                   embedding_sizes=args.embedding_sizes,
                   hidden_size=args.hidden_size,
                   nums_type=args.nums_type,
                   device = args.device
                   )

    def __init__(self, nums_type, dim_features, embedding_sizes, hidden_size, device):
        super().__init__()
        self.dim_features = dim_features
        self.embedding_sizes = embedding_sizes
        self.hidden_size = hidden_size
        self.nums_type = nums_type
        self.device = device

        # auto-encoder
        self.encodeds = [
            nn.Linear(sum([self.nums_type[j] for j in range(3) if j != i]), self.embedding_sizes[i]) for i in range(3)]
        self.decodeds = [
            nn.Linear(self.embedding_sizes[i], sum([self.nums_type[j] for j in range(3) if j != i])) for i in range(3)]
        self.hidden_layer = nn.Linear(
            sum(self.embedding_sizes), self.hidden_size)

        self.ouput_layer = nn.Linear(self.hidden_size, 1)

    def forward(self, input_ids):
        """
        The forward part of the DHNE.

        Parameters
        ----------
        input_ids :
            the input block of this batch

        Returns
        -------
        tensor
            The logits after DHNE training.
        """
        encodeds = []
        decodeds = []
        for i in range(3):
            encoded = torch.tanh(self.encodeds[i].to(self.device)(input_ids[i].to(self.device)))
            encodeds.append(encoded)
            decodeds.append(torch.sigmoid(self.decodeds[i].to(self.device)(encoded)))
        merged = torch.concat(encodeds, axis=1)
        hidden = self.hidden_layer(merged)
        hidden = torch.tanh(hidden)
        output = self.ouput_layer(hidden)
        return decodeds+[output]

    def embedding_lookup(self, index, sparse_input=False):
        return [self.embeddings[i][index[:, i], :] for i in range(3)]




import pickle
import os
from openhgnn.dataset import register_dataset, BaseDataset
import numpy as np
import copy
from torch.utils.data import Dataset
from scipy.sparse import vstack as s_vstack
from scipy.sparse import csr_matrix
from dgl.data.utils import download, extract_archive


class HgraphDataset(Dataset):

    _prefix = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/'
    _urls = {

    }

    def __init__(self, name, is_train=True, raw_dir = None, force_reload = False, verbose = True):
        assert name in ['GPS','drug','MovieLens','wordnet']

        self.data_path = './openhgnn/dataset/{}4DHNE.zip'.format(name.lower())
        raw_dir = './openhgnn/dataset/'
        url = self._prefix + 'dataset/openhgnn/{}4DHNE.zip'.format(name.lower())

        super().__init__()
        self.name=name
        self.url=url
        self.raw_dir=raw_dir
        self.force_reload=force_reload
        self.verbose=verbose
        self.train_file = 'train_data.npz'
        self.test_file = 'test_data.npz'
        self.num_neg_samples = 1
        self.pair_radio = 0.9
        self.sparse_input = True
        self.train_dir = raw_dir + '{}'.format(self.name)
        self.is_train = is_train
        self.download()
        self.process()


    def download(self):
        # download raw data to local disk
        # path to store the file
        if os.path.exists(self.data_path):  # pragma: no cover
           pass
        else:
            file_path = os.path.join(self.raw_dir)
            # download file
            download(self.url, path=file_path)
        extract_archive(self.data_path, os.path.join(self.raw_dir, self.name))

    def process(self):
        if self.is_train:
            self.process_train()
            if not os.path.exists(f"./embeddings_{self.name}.pkl"):
                self.embeddings = generate_embeddings(
                    self.edges, self.nums_type)
                with open(f"./embeddings_{self.name}.pkl", 'wb') as f:
                    pickle.dump(self.embeddings, f)
            else:
                with open(f"./embeddings_{self.name}.pkl", 'rb') as f:
                    self.embeddings = pickle.load(f)

        else:
            self.process_val()
            with open(f"./embeddings_{self.name}.pkl", 'rb') as f:
                self.embeddings = pickle.load(f)

    def process_val(self):
        data = np.load(os.path.join(self.train_dir, self.test_file), allow_pickle=True)
        self.edges = data['test_data']
        self.nums_type = data['nums_type']
        self.node_cluster = data['node_cluster'] if 'node_cluster' in data else None
        self.edge_set = set(map(tuple, self.edges))

    def process_train(self):
        data = np.load(os.path.join(self.train_dir, self.train_file), allow_pickle=True)
        self.edges = data['train_data']
        self.nums_type = data['nums_type']
        self.labels = data['labels'] if 'labels' in data else None
        self.idx_label = data['idx_label'] if 'idx_label' in data else None
        self.label_set = data['label_name'] if 'label_name' in data else None
        self.edge_set = set(map(tuple, self.edges))

    def __getitem__(self, idx):
        pos_data = [self.edges[idx]]
        neg_data = []
        n_neg = 0
        while(n_neg < self.num_neg_samples):
            # warning !!! we need deepcopy to copy list
            index = copy.deepcopy(self.edges[idx])
            mode = np.random.rand()
            if mode < self.pair_radio:
                type_ = np.random.randint(3)
                # Randomly select a mode
                node = np.random.randint(self.nums_type[type_])
                index[type_] = node
            else:
                # Randomly select two types
                types_ = np.random.choice(3, 2, replace=False)
                node_1 = np.random.randint(self.nums_type[types_[0]])
                node_2 = np.random.randint(self.nums_type[types_[1]])
                index[types_[0]] = node_1
                index[types_[1]] = node_2
            if tuple(index) in self.edge_set:
                continue
            n_neg += 1
            neg_data.append(index)
        data = np.vstack(pos_data+neg_data)
        if len(neg_data) > 0:
            nums = len(data)
            labels = np.zeros(nums)
            labels[0] = 1
        else:
            labels = np.ones(1)
        batch_e = embedding_lookup(self.embeddings, data)
        return batch_e, labels

    def __len__(self):
        return len(self.edges)


def embedding_lookup(embeddings, index, sparse_input=True):
    if sparse_input:
        return [embeddings[i][index[:, i], :].todense() for i in range(3)]
    else:
        return [embeddings[i][index[:, i], :] for i in range(3)]


def generate_H(edge, nums_type):
    nums_examples = len(edge)
    H = [csr_matrix((np.ones(nums_examples), (edge[:, i], range(
        nums_examples))), shape=(nums_type[i], nums_examples)) for i in range(3)]
    return H


def generate_embeddings(edge, nums_type):
    r"""
    Args:
        edge (_type_): Number of edges
        nums_type (_type_): Number of node types
    Returns:
        _type_: _description_
    """
    H = generate_H(edge, nums_type)
    embeddings = [H[i].dot(s_vstack([H[j] for j in range(3) if j != i]).T).astype(
        'float') for i in range(3)]
    # 0-1 scaling
    for i in range(3):
        col_max = np.array(embeddings[i].max(0).todense()).flatten()
        _, col_index = embeddings[i].nonzero()
        embeddings[i].data /= col_max[col_index]

    return embeddings


@register_dataset("hypergraph_dataset")
class HGraphDataset(BaseDataset):
    def get_data(self, name, is_train):
        return HgraphDataset(name, is_train)



import torch as th
from tqdm import tqdm
from . import BaseFlow, register_flow
from ..models import build_model
from ..models.GATNE import NSLoss
import torch
from tqdm.auto import tqdm
from numpy import random
import dgl
from ..sampler.GATNE_sampler import NeighborSampler, generate_pairs


@register_flow("GATNE_trainer")
class GATNE(BaseFlow):
    def __init__(self, args):
        super(GATNE, self).__init__(args)
        self.model = build_model(self.model).build_model_from_args(self.args, self.hg).to(self.device)

        self.train_pairs = None
        self.train_dataloader = None
        self.nsloss = None
        self.neighbor_sampler = None

        self.orig_val_hg = self.task.val_hg
        self.orig_test_hg = self.task.test_hg

        self.preprocess()
        
    def preprocess(self):
        assert len(self.hg.ntypes) == 1
        bidirected_hg = dgl.to_bidirected(dgl.to_simple(self.hg.to('cpu')))
        all_walks = []
        for etype in self.hg.etypes:
            nodes = torch.unique(bidirected_hg.edges(etype=etype)[0]).repeat(self.args.rw_walks)
            traces, types = dgl.sampling.random_walk(
                bidirected_hg, nodes, metapath=[etype] * (self.args.rw_length - 1)
            )
            all_walks.append(traces)
        self.train_pairs = generate_pairs(all_walks, self.args.window_size, self.args.num_workers)
        self.neighbor_sampler = NeighborSampler(bidirected_hg, [self.args.neighbor_samples])
        self.train_dataloader = torch.utils.data.DataLoader(
            self.train_pairs,
            batch_size=self.args.batch_size,
            collate_fn=self.neighbor_sampler.sample,
            shuffle=True,
            num_workers=self.args.num_workers,
            pin_memory=True,
        )
        self.nsloss = NSLoss(self.hg.num_nodes(), self.args.neg_size, self.args.dim).to(self.device)
        self.optimizer = torch.optim.Adam(
            [{"params": self.model.parameters()}, {"params": self.nsloss.parameters()}], lr=self.args.learning_rate
        )
        return

    def train(self):
        best_score = 0
        patience = 0
        for self.epoch in range(self.args.max_epoch):
            self._full_train_step()
            cur_score = self._full_test_step()
            if cur_score > best_score:
                best_score = cur_score
                patience = 0
            else:
                patience += 1
                if patience > self.args.patience:
                    self.logger.train_info(f'Early Stop!\tEpoch:{self.epoch:03d}.')
                    break

    def _full_train_step(self):
        self.model.train()
        random.shuffle(self.train_pairs)
        data_iter = tqdm(
            self.train_dataloader,
            desc="epoch %d" % self.epoch,
            total=(len(self.train_pairs) + (self.args.batch_size - 1)) // self.args.batch_size,
        )
        avg_loss = 0.0

        for i, (block, head_invmap, tails, block_types) in enumerate(data_iter):
            self.optimizer.zero_grad()
            # embs: [batch_size, edge_type_count, embedding_size]
            block_types = block_types.to(self.device)
            embs = self.model(block[0].to(self.device))[head_invmap]
            embs = embs.gather(
                1, block_types.view(-1, 1, 1).expand(embs.shape[0], 1, embs.shape[2])
            )[:, 0]
            loss = self.nsloss(
                block[0].dstdata[dgl.NID][head_invmap].to(self.device),
                embs,
                tails.to(self.device),
            )
            loss.backward()
            self.optimizer.step()
            avg_loss += loss.item()

            post_fix = {
                "epoch": self.epoch,
                "iter": i,
                "avg_loss": avg_loss / (i + 1),
                "loss": loss.item(),
            }
            data_iter.set_postfix(post_fix)

    def _full_test_step(self):
        self.model.eval()
        # {'1': {}, '2': {}}
        final_model = dict(
            zip(self.hg.etypes, [th.empty(self.hg.num_nodes(), self.args.dim) for _ in range(len(self.hg.etypes))]))

        for i in tqdm(range(self.hg.num_nodes()), desc='Evaluating...'):
            train_inputs = (
                torch.tensor([i for _ in range(len(self.hg.etypes))])
                    .unsqueeze(1)
                    .to(self.device)
            )  # [i, i]
            train_types = (
                torch.tensor(list(range(len(self.hg.etypes)))).unsqueeze(1).to(self.device)
            )  # [0, 1]
            pairs = torch.cat(
                (train_inputs, train_inputs, train_types), dim=1
            )  # (2, 3)
            (
                train_blocks,
                train_invmap,
                fake_tails,
                train_types,
            ) = self.neighbor_sampler.sample(pairs)

            node_emb = self.model(train_blocks[0].to(self.device))[train_invmap]
            node_emb = node_emb.gather(
                1,
                train_types.to(self.device)
                    .view(-1, 1, 1)
                    .expand(node_emb.shape[0], 1, node_emb.shape[2]),
            )[:, 0]

            for j in range(len(self.hg.etypes)):
                final_model[self.hg.etypes[j]][i] = node_emb[j].detach()
        metric = {}
        score = []
        for etype in self.hg.etypes:
            self.task.val_hg = dgl.edge_type_subgraph(self.orig_val_hg, [etype])
            self.task.test_hg = dgl.edge_type_subgraph(self.orig_test_hg, [etype])

            for split in ['test', 'valid']:
                n_embedding = {self.hg.ntypes[0]: final_model[etype].to(self.device)}
                res = self.task.evaluate(n_embedding=n_embedding, mode=split)
                metric[split] = res
                if split == 'valid':
                    score.append(res.get('roc_auc'))
            self.logger.train_info(etype + self.logger.metric2str(metric))

        avg_score = sum(score) / len(score)
        return avg_score



import torch as t
from dgl.data import DGLDataset
from dgl.data.utils import download, extract_archive
from dgl.data.utils import load_graphs
import os
import numpy as np
import dgl
import pickle


class HGCLDataset(DGLDataset):

    _prefix = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/'
    _urls = {

    }

    def __init__(self, name, raw_dir=None, force_reload=False, verbose=True):
        assert name in ['Epinions', 'CiaoDVD', 'Yelp']
        self.data_path = './{}.zip'.format(name)
        self.g_path = './{}/graph.bin'.format(name)
        raw_dir = './'
        url = 'https://s3.cn-north-1.amazonaws.com.cn/dgl-data/' + 'dataset/{}.zip'.format(name)

        super(HGCLDataset, self).__init__(name=name,
                                          url=url,
                                        raw_dir=raw_dir,
                                        force_reload=force_reload,
                                        verbose=verbose)
    def create_graph(self):
        '''
            raw_dataset url : https://drive.google.com/drive/folders/1s6LGibPnal6gMld5t63aK4J7hnVkNeDs
        '''
        data_path = self.data_path + '/data.pkl'
        distance_path = self.data_path + '/distanceMat_addIUUI.pkl'
        ici_path = self.data_path + '/ICI.pkl'

        with open(data_path, 'rb') as fs:
            data = pickle.load(fs)
        with open(distance_path, 'rb') as fs:
            distanceMat = pickle.load(fs)
        with open(ici_path, "rb") as fs:
            itemMat = pickle.load(fs)

        trainMat, testdata, _, categoryMat, _ = data
        userNum, itemNum = trainMat.shape
        userDistanceMat, itemDistanceMat, uiDistanceMat = distanceMat

        # trainMat
        trainMat_coo = trainMat.tocoo()
        trainMat_i, trainMat_j, trainMat_data = trainMat_coo.row, trainMat_coo.col, trainMat_coo.data

        # testdata
        testdata = np.array(testdata)

        # userDistanceMat
        userDistanceMat_coo = userDistanceMat.tocoo()
        userDistanceMat_i, userDistanceMat_j, userDistanceMat_data = userDistanceMat_coo.row, userDistanceMat_coo.col, userDistanceMat_coo.data

        # itemMat
        itemMat_coo = itemMat.tocoo()
        itemMat_i, itemMat_j, itemMat_data = itemMat_coo.row, itemMat_coo.col, itemMat_coo.data

        # uiDisantanceMat
        uiDistanceMat_coo = uiDistanceMat.tocoo()
        uiDistanceMat_i, uiDistanceMat_j, uiDistanceMat_data = uiDistanceMat_coo.row, uiDistanceMat_coo.col, uiDistanceMat_coo.data

        graph_data = {
            ('user', 'interact_train', 'item'): (t.tensor(trainMat_i), t.tensor(trainMat_j)),
            ('user', 'distance', 'user'): (t.tensor(userDistanceMat_i), t.tensor(userDistanceMat_j)),
            ('item', 'distance', 'item'): (t.tensor(itemMat_i), t.tensor(itemMat_j)),
            ('user+item', 'distance', 'user+item'): (t.tensor(uiDistanceMat_i), t.tensor(uiDistanceMat_j)),
            ('user', 'interact_test', 'item'): (t.tensor(testdata[:, 0]), t.tensor(testdata[:, 1]))
        }
        g = dgl.heterograph(graph_data)
        dgl.save_graphs(self.data_path + '/graph.bin', g)
        self.g_path = self.data_path + '/graph.bin'


    def download(self):
        # download raw data to local disk
        # path to store the file
        if os.path.exists(self.data_path):  # pragma: no cover
           pass
        else:
            file_path = os.path.join(self.raw_dir)
            # download file
            download(self.url, path=file_path)
        extract_archive(self.data_path, os.path.join(self.raw_dir, self.name))

    def process(self):
        # process raw data to graphs, labels, splitting masks
        g, _ = load_graphs(self.g_path)
        self._g = g

    def __getitem__(self, idx):
        # get one example by index
        return self._g[idx]

    def __len__(self):
        # number of data examples
        return 1

    def save(self):
        # save processed data to directory `self.save_path`
        pass

    def load(self):
        # load processed data from directory `self.save_path`
        pass

    def has_cache(self):
        # check whether there are processed data in `self.save_path`
        pass



import importlib
from .base_flow import BaseFlow

FLOW_REGISTRY = {}


def register_flow(name):
    """
    New flow can be added to openhgnn with the :func:`register_flow`
    function decorator.

    For example::

        @register_task('node_classification')
        class NodeClassification(BaseFlow):
            (...)

    Args:
        name (str): the name of the flows
    """

    def register_flow_cls(cls):
        if name in FLOW_REGISTRY:
            raise ValueError("Cannot register duplicate flow ({})".format(name))
        if not issubclass(cls, BaseFlow):
            raise ValueError("Flow ({}: {}) must extend BaseFlow".format(name, cls.__name__))
        FLOW_REGISTRY[name] = cls
        return cls

    return register_flow_cls


def try_import_flow(flow):
    if flow not in FLOW_REGISTRY:
        if flow in SUPPORTED_FLOWS:
            importlib.import_module(SUPPORTED_FLOWS[flow])
        else:
            print(f"Failed to import {flow} flows.")
            return False
    return True


def build_flow(args, flow_name):
    if not try_import_flow(flow_name):
        exit(1)
    return FLOW_REGISTRY[flow_name](args)


SUPPORTED_FLOWS = {
    'entity_classification': 'openhgnn.trainerflow.entity_classification',
    'node_classification': 'openhgnn.trainerflow.node_classification',
    'node_classification_ac': 'openhgnn.trainerflow.node_classfication_ac',
    'distmult': 'openhgnn.trainerflow.dist_mult',
    'link_prediction': 'openhgnn.trainerflow.link_prediction',
    'recommendation': 'openhgnn.trainerflow.recommendation',
    'hetgnntrainer': 'openhgnn.trainerflow.hetgnn_trainer',
    'hgttrainer': 'openhgnn.trainerflow.hgt_trainer',
    'nshetrainer': 'openhgnn.trainerflow.nshe_trainer',
    'demo': 'openhgnn.trainerflow.demo',
    'kgcntrainer': 'openhgnn.trainerflow.kgcn_trainer',
    'HeGAN_trainer': 'openhgnn.trainerflow.HeGAN_trainer',
    'mp2vec_trainer': 'openhgnn.trainerflow.mp2vec_trainer',
    'herec_trainer': 'openhgnn.trainerflow.herec_trainer',
    'HeCo_trainer': 'openhgnn.trainerflow.HeCo_trainer',
    'DMGI_trainer': 'openhgnn.trainerflow.DMGI_trainer',
    'slicetrainer': 'openhgnn.trainerflow.slice_trainer',
    'hde_trainer': 'openhgnn.trainerflow.hde_trainer',
    'GATNE_trainer': 'openhgnn.trainerflow.GATNE_trainer',
    'TransX_trainer': 'openhgnn.trainerflow.TransX_trainer',
    'han_nc_trainer': 'openhgnn.trainerflow.HANNodeClassification',
    'han_lp_trainer': 'openhgnn.trainerflow.HANLinkPrediction',
    'RoHe_trainer': 'openhgnn.trainerflow.RoHe_trainer',
    'mg2vec_trainer': 'openhgnn.trainerflow.mg2vec_trainer',
    'DHNE_trainer': 'openhgnn.trainerflow.DHNE_trainer',
    'DiffMG_trainer': 'openhgnn.trainerflow.DiffMG_trainer',
    'MeiREC_trainer': 'openhgnn.trainerflow.MeiRec_trainer',
    'abnorm_event_detection': 'openhgnn.trainerflow.AbnormEventDetection',
    'SHGP_trainer': 'openhgnn.trainerflow.SHGP_trainer',
    'KGAT_trainer': 'openhgnn.trainerflow.KGAT_trainer',
    'DSSL_trainer': 'openhgnn.trainerflow.DSSL_trainer',
    'hgcltrainer': 'openhgnn.trainerflow.hgcl_trainer'
}

from .hgcl_trainer import HGCLtrainer
from .node_classification import NodeClassification
from .link_prediction import LinkPrediction
from .recommendation import Recommendation
from .hetgnn_trainer import HetGNNTrainer
from .hgt_trainer import HGTTrainer
from .kgcn_trainer import KGCNTrainer
from .HeGAN_trainer import HeGANTrainer
from .mp2vec_trainer import Metapath2VecTrainer
from .herec_trainer import HERecTrainer
from .HeCo_trainer import HeCoTrainer
from .DMGI_trainer import DMGI_trainer
from .slice_trainer import SLiCETrainer
from .hde_trainer import hde_trainer
from .GATNE_trainer import GATNE
from .han_trainer import HANNodeClassification
from .han_trainer import HANLinkPrediction
from .RoHe_trainer import RoHeTrainer
from .mg2vec_trainer import Mg2vecTrainer
from .DHNE_trainer import DHNE_trainer
from .DiffMG_trainer import DiffMG_trainer
from .MeiRec_trainer import MeiRECTrainer
from .kgat_trainer import KGAT_Trainer
from .node_classification_ac import NodeClassificationAC
from .DSSL_trainer import DSSL_trainer

__all__ = [
    'BaseFlow',
    'NodeClassification',
    'LinkPrediction',
    'Recommendation',
    'HetGNNTrainer',
    'HGTTrainer',
    'KGCNTrainer',
    'HeGANTrainer',
    'Metapath2VecTrainer',
    'HERecTrainer',
    'HeCoTrainer',
    'DMGI_trainer',
    'SLiCETrainer',
    'hde_trainer',
    'GATNE',
    'HANNodeClassification',
    'HANLinkPrediction',
    'Mg2vecTrainer',
    'DHNE_trainer',
    'DiffMG_trainer',
    'MeiRECTrainer',
<<<<<<< HEAD
    'KGAT_Trainer',
    'DSSL_trainer'
=======
    'HGCLtrainer'
>>>>>>> a8f8928e3ae66f22a19b31c7a8e95b9db2643b5b
]
classes = __all__



import torch.nn.functional as F
import torch.nn as nn
from openhgnn.tasks import BaseTask, register_task
from openhgnn.utils import Evaluator
from dgl.data import DGLDataset
from openhgnn.dataset import build_dataset, CLASS_DATASETS, DATASET_REGISTRY, build_dataset_v2, try_import_task_dataset
import torch


@register_task("hypergraph")
class hypergraph(BaseTask):
    r"""
    Hypergraph tasks.

    Attributes
    -----------
    dataset : HypergraphDataset
        Task-related dataset

    evaluator : Evaluator
        offer evaluation metric


    Methods
    ---------
    get_graph :
        return a graph
    get_loss_fn :
        return a loss function
    """

    def __init__(self, args):
        super(hypergraph, self).__init__()
        self.logger = args.logger
        self.dataset = build_dataset(
            args.dataset, 'hypergraph_dataset', logger=self.logger)
        # self.evaluator = Evaluator()
        self.logger = args.logger
        self.train_data = self.dataset.get_data(args.dataset, is_train=True)
        self.val_data = self.dataset.get_data(args.dataset, is_train=False)

        self.evaluator = Evaluator(args.seed)

    def get_graph(self):
        return torch.tensor(0)

    def get_loss_fn(self):
        return F.binary_cross_entropy

    def get_evaluator(self, name):
        if name == 'acc':
            return self.evaluator.c
            al_acc
        elif name == 'f1_lr':
            return self.evaluator.nc_with_LR
        elif name == 'f1':
            return self.evaluator.f1_node_classification

    def evaluate(self, logits, mode='test', info=True):
        r"""

        Parameters
        ----------
        logits: th.Tensor
            the logits of batch
        mode: str
            the evaluation mode, train/valid/test
        Returns
        -------

        """
        if mode == 'test':
            mask = self.test_idx
        elif mode == 'valid':
            mask = self.val_idx
        elif mode == 'train':
            mask = self.train_idx

        if self.multi_label:
            pred = (logits[mask].cpu().numpy() > 0).astype(int)
        else:
            pred = logits[mask].argmax(dim=1).to('cpu')

        if self.evaluation_metric == 'acc':
            acc = self.evaluator.cal_acc(self.labels[mask], pred)
            return dict(Accuracy=acc)
        elif self.evaluation_metric == 'acc-ogbn-mag':
            from ogb.nodeproppred import Evaluator
            evaluator = Evaluator(name='ogbn-mag')
            logits = logits.unsqueeze(dim=1)
            input_dict = {"y_true": logits,
                          "y_pred": self.labels[self.test_idx]}
            result_dict = evaluator.eval(input_dict)
            return result_dict
        elif self.evaluation_metric == 'f1':
            f1_dict = self.evaluator.f1_node_classification(
                self.labels[mask], pred)
            return f1_dict
        else:
            raise ValueError('The evaluation metric is not supported!')



# Read and process the meta.txt

from torch.utils.data import Dataset

import random
import math


class DataReader:
    def __init__(self, filename, block_size):
        self.filename = filename
        self.node_count = 0
        self.mg_count = 0
        self.node_dict = dict()
        self.mg_dict = dict()
        self.mg_num_dict = dict()
        self.node_reverse_dict = dict()
        self.mg_reverse_dict = dict()
        self.unigram = list()

        self.data_index = 0
        self.epoch_end = False

        self.file = open(filename, 'r')
        self.block_size = block_size
        self.train_file = list()

        self.read_train_file()

    def read_train_file(self):
        cnt = 0
        with open(self.filename, 'r') as f:
            for line in f:
                temp = list(line.strip('\n').split())
                if temp[2][1:] not in self.mg_dict:
                    self.mg_dict[temp[2][1:]] = self.mg_count
                    self.mg_num_dict[temp[2][1:]] = float(temp[3][1:])
                    self.mg_count += 1
                else:
                    self.mg_num_dict[temp[2][1:]] += float(temp[3][1:])
                if temp[0] not in self.node_dict:
                    self.node_dict[temp[0]] = self.node_count
                    self.node_count += 1
                if temp[1] not in self.node_dict:
                    self.node_dict[temp[1]] = self.node_count
                    self.node_count += 1
                cnt += 1
                if cnt % 10000000 == 0:
                    print("read lines: ", cnt)
        self.node_reverse_dict = dict(zip(self.node_dict.values(), self.node_dict.keys()))
        self.mg_reverse_dict = dict(zip(self.mg_dict.values(), self.mg_dict.keys()))
        for i in range(self.mg_count):
            self.unigram.append(int(math.ceil(self.mg_num_dict[self.mg_reverse_dict[i]])))
        print("read finish, total lines: ", cnt)
        print("total nodes: ", self.node_count)
        print("total metaGraphs: ", self.mg_count)

    def read_block(self):
        self.train_file = [0] * self.block_size
        instance_count = 0
        for i in range(self.block_size):
            line = self.file.readline()
            if line:
                temp = list(line.strip('\n').split())
                temp[0] = self.node_dict[temp[0]]
                temp[1] = self.node_dict[temp[1]]
                temp[2] = self.mg_dict[temp[2][1:]]
                temp[3] = float(temp[3][1:])
                self.train_file[instance_count] = temp
                instance_count += 1
            else:
                self.file.seek(0, 0)
                self.epoch_end = True
                break
        self.train_file = self.train_file[:instance_count]
        random.shuffle(self.train_file)


class Mg2vecSampler(Dataset):
    def __init__(self, filename, block_size, alpha):
        self.data = DataReader(filename, block_size)
        self.alpha = alpha
        self.data.read_block()

    def __len__(self):
        return len(self.data.train_file)

    def __getitem__(self, index):
        a = int(self.data.train_file[index][0])
        b = int(self.data.train_file[index][1])
        weight = (1 - self.alpha) if a == b else self.alpha
        label = int(self.data.train_file[index][2])
        freq = self.data.train_file[index][3]
        return a, b, label, freq, weight



import dgl
import torch
import torch.nn as nn
import dgl.function as Fn
import torch.nn.functional as F

from dgl.ops import edge_softmax
from dgl.nn.pytorch import HeteroLinear
from . import BaseModel, register_model
from ..utils import to_hetero_feat

@register_model('HGAT')
class HGAT(BaseModel):
    r"""
    This is a model HGAT from `Heterogeneous Graph Attention Networks for Semi-supervised Short Text Classification
    <https://dl.acm.org/doi/abs/10.1145/3450352>`__

    It contains the following parts:

    Type-level Attention: Given a specific node :math:`v`, we need to calculate the type-level attention scores based on the current node 
    embedding and the type embedding.
    
    .. math::
       a_{\tau} = \sigma(\mu_{\tau}^T \cdot [h_v \parallel h_{\tau}]) \quad (1)
    
    The type embedding is :math:`h_{\tau}=\sum_{v^{'}}\widetilde{A}_{vv^{'}}h_{v^{'}}`, 
    which is the sum of the neighboring node features :math:`h_{v^{'}}` 
    where the nodes :math:`v^{'} \in \mathcal{N}_v` and are with the type :math:`h_{\tau}`.
    :math:`\mu_{\tau}` is the attention vector for the type :math:`\tau`.
    
    And the type-level attention weights is:
    
    .. math::
       \alpha_{\tau} = \frac{exp(a_{\tau})}{\sum_{\tau^{'}\in \mathcal{T}} exp(a_{\tau^{'}})} \quad (2)

    Node-level Attention: Given a specific node :math:`v` and its neightoring node :math:`v^{'}\in \mathcal{N}_v`, 
    we need to calculate the node-level attention scores based on the node embeddings :math:`h_v` and :math:`h_{v^{'}}`
    and with the type-level attention weight :math:`\alpha_{\tau^{'}}` for the node :math:`v^{'}`:
    
    .. math::
       b_{vv^{'}} = \sigma(\nu^T \cdot \alpha_{\tau^{'}}[h_v \parallel h_{v^{'}}]) \quad (3)
    
    where :math:`\nu` is the attention vector.
    
    And the node-level attention weights is:
    
    .. math::
       \beta_{vv^{'}} = \frac{exp(b_{vv^{'}})}{\sum_{i\in \mathcal{N}_v} exp(b_{vi})} \quad (4)
    
    The final output is:
    
    .. math::
       H^{(l+1)} = \sigma(\sum_{\tau \in \mathcal{T}}B_{\tau}\cdot H_{\tau}^{(l)}\cdot W_{\tau}^{(l)}) \quad (5)
    
    Parameters
    ----------
    num_layers: int
        the number of layers we used in the computing
    in_dim: int
        the input dimension
    hidden_dim: int
        the hidden dimension
    num_classes: int
        the number of the output classes
    ntypes: list
        the list of the node type in the graph
    negative_slope: float
        the negative slope used in the LeakyReLU
    """
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(args.num_layers,
                   args.hidden_dim,
                   args.num_classes,
                   hg.ntypes,
                   args.negative_slope)
    
    def __init__(self, num_layers, hidden_dim,
                 num_classes, ntypes, negative_slope):
        super(HGAT, self).__init__()
        self.num_layers = num_layers
        self.activation = F.elu
        
        
        self.hgat_layers = nn.ModuleList()
        self.hgat_layers.append(
            TypeAttention(hidden_dim,
                          ntypes,
                          negative_slope))
        self.hgat_layers.append(
            NodeAttention(hidden_dim,
                          hidden_dim,
                          negative_slope)
        )
        for l in range(num_layers - 1):
            self.hgat_layers.append(
                TypeAttention(hidden_dim,
                            ntypes,
                            negative_slope))
            self.hgat_layers.append(
                NodeAttention(hidden_dim,
                            hidden_dim,
                            negative_slope)
            )
        
        self.hgat_layers.append(
            TypeAttention(hidden_dim,
                          ntypes,
                          negative_slope))
        self.hgat_layers.append(
            NodeAttention(hidden_dim,
                          num_classes,
                          negative_slope)
        )
        
        
    def forward(self, hg, h_dict):
        """
        The forward part of the HGAT.
        
        Parameters
        ----------
        hg : object
            the dgl heterogeneous graph
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after the output projection.
        """
        with hg.local_scope():
            hg.ndata['h'] = h_dict
            for l in range(self.num_layers):
                attention = self.hgat_layers[2 * l](hg, hg.ndata['h'])
                hg.edata['alpha'] = attention
                g = dgl.to_homogeneous(hg, ndata = 'h', edata = ['alpha'])
                h = self.hgat_layers[2 * l + 1](g, g.ndata['h'], g.ndata['_TYPE'], g.ndata['_TYPE'], presorted = True)
                h_dict = to_hetero_feat(h, g.ndata['_TYPE'], hg.ntypes)
                hg.ndata['h'] = h_dict

        return h_dict

class TypeAttention(nn.Module):
    """
    The type-level attention layer

    Parameters
    ----------
    in_dim: int
        the input dimension of the feature
    ntypes: list
        the list of the node type in the graph
    slope: float
        the negative slope used in the LeakyReLU
    """
    def __init__(self, in_dim, ntypes, slope):
        super(TypeAttention, self).__init__()
        attn_vector = {}
        for ntype in ntypes:
            attn_vector[ntype] = in_dim
        self.mu_l = HeteroLinear(attn_vector, in_dim)
        self.mu_r = HeteroLinear(attn_vector, in_dim)
        self.leakyrelu = nn.LeakyReLU(slope)
        
    def forward(self, hg, h_dict):
        """
        The forward part of the TypeAttention.
        
        Parameters
        ----------
        hg : object
            the dgl heterogeneous graph
        h_dict: dict
            the feature dict of different node types
            
        Returns
        -------
        dict
            The embeddings after the output projection.
        """
        h_t = {}
        attention = {}
        with hg.local_scope():
            hg.ndata['h'] = h_dict
            for srctype, etype, dsttype in hg.canonical_etypes:
                rel_graph = hg[srctype, etype, dsttype]
                if srctype not in h_dict:
                    continue
                with rel_graph.local_scope():
                    degs = rel_graph.out_degrees().float().clamp(min = 1)
                    norm = torch.pow(degs, -0.5)
                    feat_src = h_dict[srctype]
                    shp = norm.shape + (1,) * (feat_src.dim() - 1)
                    norm = torch.reshape(norm, shp)
                    feat_src = feat_src * norm
                    rel_graph.srcdata['h'] = feat_src
                    rel_graph.update_all(Fn.copy_u('h', 'm'), Fn.sum(msg='m', out='h'))
                    rst = rel_graph.dstdata['h']
                    degs = rel_graph.in_degrees().float().clamp(min=1)
                    norm = torch.pow(degs, -0.5)
                    shp = norm.shape + (1,) * (feat_src.dim() - 1)
                    norm = torch.reshape(norm, shp)
                    rst = rst * norm
                    h_t[srctype] = rst
                    h_l = self.mu_l(h_dict)[dsttype]
                    h_r = self.mu_r(h_t)[srctype]
                    edge_attention = F.elu(h_l + h_r)
                    # edge_attention = F.elu(h_l + h_r).unsqueeze(0)
                    rel_graph.ndata['m'] = {dsttype: edge_attention,
                                    srctype: torch.zeros((rel_graph.num_nodes(ntype = srctype),)).to(edge_attention.device)}
                    # print(rel_graph.ndata)
                    reverse_graph = dgl.reverse(rel_graph)
                    reverse_graph.apply_edges(Fn.copy_u('m', 'alpha'))
                
                hg.edata['alpha'] = {(srctype, etype, dsttype): reverse_graph.edata['alpha']}
                
                # if dsttype not in attention.keys():
                #     attention[dsttype] = edge_attention
                # else:
                #     attention[dsttype] = torch.cat((attention[dsttype], edge_attention))
            attention = edge_softmax(hg, hg.edata['alpha'])
            # for ntype in hg.dsttypes:
            #     attention[ntype] = F.softmax(attention[ntype], dim = 0)

        return attention
    
class NodeAttention(nn.Module):
    """
    The node-level attention layer

    Parameters
    ----------
    in_dim: int
        the input dimension of the feature
    out_dim: int
        the output dimension
    slope: float
        the negative slope used in the LeakyReLU
    """
    def __init__(self, in_dim, out_dim, slope):
        super(NodeAttention, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.Mu_l = nn.Linear(in_dim, in_dim)
        self.Mu_r = nn.Linear(in_dim, in_dim)
        self.leakyrelu = nn.LeakyReLU(slope)
        
    def forward(self, g, x, ntype, etype, presorted = False):
        """
        The forward part of the NodeAttention.

        Parameters
        ----------
        g : object
            the dgl homogeneous graph
        x: tensor
            the original features of the graph
        ntype: tensor
            the node type of the graph
        etype: tensor
            the edge type of the graph
        presorted: boolean
            if the ntype and etype are preordered, default: ``False``
            
        Returns
        -------
        tensor
            The embeddings after aggregation.
        """
        with g.local_scope():
            src = g.edges()[0]
            dst = g.edges()[1]
            h_l = self.Mu_l(x)[src]
            h_r = self.Mu_r(x)[dst]
            edge_attention = self.leakyrelu((h_l + h_r) * g.edata['alpha'])
            edge_attention = edge_softmax(g, edge_attention)
            g.edata['alpha'] = edge_attention
            g.srcdata['x'] = x
            g.update_all(Fn.u_mul_e('x', 'alpha', 'm'),
                         Fn.sum('m', 'x'))
            h = g.ndata['x']
        return h


import torch.nn as nn
import torch.nn.functional as F
import dgl.nn as dglnn
from openhgnn.models import BaseModel, register_model
from dgl.nn.pytorch.conv import GINConv

@register_model('GIN')
class GIN(BaseModel):
    @classmethod
    def build_model_from_args(cls, args, hg):
        return cls(input_dim=args.input_dim,
                   hidden_dim=args.hidden_dim,
                   output_dim=args.out_dim,
                   num_hidden_layers=args.num_layers - 1,
                   rel_names=hg.etypes,
                   learn_eps=args.learn_eps,
                   aggregate=args.aggregate)

    def __init__(self, input_dim, hidden_dim, output_dim, num_hidden_layers, rel_names, learn_eps, aggregate):
        super(GIN, self).__init__()
        self.rel_names = rel_names
        self.learn_eps = learn_eps
        self.aggregate = aggregate
        self.layers = nn.ModuleList()
        # input 2 hidden
        self.layers.append(GINLayer(
            input_dim, hidden_dim, self.rel_names, self.learn_eps, self.aggregate))
        for i in range(num_hidden_layers):
            self.layers.append(GINLayer(
                hidden_dim, hidden_dim, self.rel_names, self.learn_eps, self.aggregate
            ))
        self.linear_prediction = nn.ModuleList()
        for _ in range(num_hidden_layers + 1):
            self.linear_prediction.append(nn.Linear(hidden_dim, output_dim))
        self.drop = nn.Dropout(0.5)

    def forward(self, hg, h_dict):
        output_dict = dict()
        logits = dict()
        
        if hasattr(hg, 'ntypes'):
            # full graph training,
            for ntype in hg.ntypes:
                output_dict[ntype] = []
                logits[ntype] = 0
            # Perform forward pass through all layers of the GNN
            for layer in self.layers:
                h_dict = layer(hg, h_dict)
                for ntype in hg.ntypes:
                    output_dict[ntype].append(h_dict[ntype])
            # perform graph sum pooling over all nodes in each layer
            for ntype in hg.ntypes:
                for i, h in enumerate(output_dict[ntype]):
                    logits[ntype] = logits[ntype] + self.drop(self.linear_prediction[i](h))
                logits[ntype] = F.softmax(logits[ntype], dim=-1)
                
        else:
            # minibatch training, block
            for layer, block in zip(self.layers, hg):
                for ntype in block.ntypes:
                    output_dict[ntype] = []
                    logits[ntype] = 0
                # Perform forward pass through all layers of the GNN
                h_dict = layer(block, h_dict)
                for ntype in block.ntypes:
                    output_dict[ntype].append(h_dict[ntype])
            # perform graph sum pooling over all nodes in each layer
            for ntype in block.ntypes:
                for i, h in enumerate(output_dict[ntype]):
                    logits[ntype] = logits[ntype] + self.drop(self.linear_prediction[i](h))
                logits[ntype] = F.softmax(logits[ntype], dim=-1)
        
        return logits

class GINLayer(nn.Module):
    def __init__(self, input_dim, output_dim, rel_names, learn_eps, aggregate):
        super(GINLayer, self).__init__()
        self.conv = dglnn.HeteroGraphConv({
            rel: GINBase(input_dim, output_dim, learn_eps)
            for rel in rel_names
        }, aggregate)

    def forward(self, g, h_dict):
        h_dict = self.conv(g, h_dict)
        out_put = {}
        for n_type, h in h_dict.items():
            out_put[n_type] = h.squeeze()
        return out_put

class GINBase(nn.Module):
    def __init__(self, input_dim, output_dim, learn_eps):
        super(GINBase, self).__init__()

        mlp = MLP(input_dim, output_dim)
        self.ginlayer = GINConv(mlp, learn_eps=learn_eps) # set to True if learning epsilon
        self.batch_norm = nn.BatchNorm1d(output_dim)

    def forward(self, g, h):
        h = self.ginlayer(g, h)
        h = self.batch_norm(h)
        h = F.relu(h)
        return h

class MLP(nn.Module):
    """Construct two-layer MLP-type aggreator for GIN model"""

    def __init__(self, input_dim, output_dim):
        super(MLP, self).__init__()
        self.linears = nn.ModuleList()
        # two-layer MLP
        self.linears.append(nn.Linear(input_dim, output_dim, bias=False))
        self.linears.append(nn.Linear(output_dim, output_dim, bias=False))
        self.batch_norm = nn.BatchNorm1d((output_dim))

    def forward(self, x):
        h = x
        h = F.relu(self.batch_norm(self.linears[0](h)))
        return self.linears[1](h)
